# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
from typing import Iterable, Tuple

import numpy as np
from wulfric import TORADIANS, absolute_to_relative

from magnopy.spinham.hamiltonian import SpinHamiltonian
from magnopy.units.inside import ENERGY
from magnopy.units.si import BOHR_MAGNETON

_logger = logging.getLogger(__name__)

__all__ = ["Energy"]

# Convert to the internal units of energy
BOHR_MAGNETON /= ENERGY

################################################################################
#                  Functions that check or make initial guess                  #
#                                                                              #
# The idea is to call it in the generalized way and obtain a vector for the    #
# minimization                                                                 #
#                                                                              #
#                   _starting(n_spins, initial_guess) -> x                     #
#                                                                              #
# Nature of the returned x depends on the type of the ground state (as well    #
# as nature of initial_guess), but nature of x has to match the one of the     #
# corresponding _update function                                               #
################################################################################


def _starting_ferro(n_spins, initial_guess=None):
    # Construct or check initial guess
    if initial_guess is None:
        spin_orientation = np.random.random((n_spins, 3))
    else:
        # Check that the input is correct
        try:
            spin_orientation = np.array(initial_guess, dtype=float)
        except:
            raise ValueError(f"spin_orientation is not array-like: {spin_orientation}")

        if spin_orientation.shape[-1] != 3:
            raise ValueError(
                f"spin_orientation has to have the last dimension of "
                f"size 3, got {spin_orientation.shape[-1]}"
            )

        # Make the input array have (I,3) shape for I = 1
        if n_spins == 1 and spin_orientation.shape == (3,):
            spin_orientation = np.array([spin_orientation])

        if n_spins != spin_orientation.shape[0]:
            raise ValueError(
                f"Expected {n_spins} spins in initial guess, "
                f"got {spin_orientation.shape[0]}"
            )

    for i in range(n_spins):
        spin_orientation[i] /= np.linalg.norm(spin_orientation[i])

    return spin_orientation


def _starting_antiferro(n_spins, antiferro_q, initial_guess=None):
    if initial_guess is not None and len(initial_guess) != 2:
        raise ValueError(
            "If initial guess is not None, then tuple with two elements "
            "is expected: (spin_orientation, cone_axis), "
            f"got {len(initial_guess)} elements."
        )

    # Get spin_orientation
    if initial_guess is None or initial_guess[0] is None:
        spin_orientation = _starting_ferro(n_spins)
    else:
        spin_orientation = _starting_ferro(n_spins, initial_guess[0])

    # Get cone_axis
    if initial_guess is None or initial_guess[1] is None:
        cone_axis = np.random.random(3)
    else:
        # Check that the input is correct
        try:
            cone_axis = np.array(initial_guess[1], dtype=float)
        except:
            raise ValueError(
                f"Initial guess for cone_axis is not array-like: " f"{initial_guess[1]}"
            )
        if cone_axis.shape != (3,):
            raise ValueError(
                f"Initial guess for cone_axis should have shape "
                f"of (3,), got: {cone_axis.shape}"
            )

    # Normalize cone axis:
    cone_axis /= np.linalg.norm(cone_axis)

    # Check that the input is correct
    try:
        spiral_vector = np.array(antiferro_q, dtype=float)
    except:
        raise ValueError(f"Spiral_vector is not array-like: " f"{antiferro_q}")
    if spiral_vector.shape != (3,):
        raise ValueError(
            f"Spiral_vector should have shape " f"of (3,), got: {spiral_vector.shape}"
        )

    return [np.concatenate((spin_orientation, [cone_axis]), axis=0), spiral_vector]


def _starting_spiral(n_spins, initial_guess=None):
    if initial_guess is not None and len(initial_guess) != 3:
        raise ValueError(
            "If initial guess is not None, then tuple with three elements "
            "is expected: (spin_orientation, cone_axis, spiral_vector), "
            f"got {len(initial_guess)} elements."
        )

    # Get spin_orientation
    if initial_guess is None or initial_guess[0] is None:
        spin_orientation = _starting_ferro(n_spins)
    else:
        spin_orientation = _starting_ferro(n_spins, initial_guess[0])

    # Get cone_axis
    if initial_guess is None or initial_guess[1] is None:
        cone_axis = np.random.random(3)
    else:
        # Check that the input is correct
        try:
            cone_axis = np.array(initial_guess[1], dtype=float)
        except:
            raise ValueError(
                f"Initial guess for cone_axis is not array-like: " f"{initial_guess[1]}"
            )
        if cone_axis.shape != (3,):
            raise ValueError(
                f"Initial guess for cone_axis should have shape "
                f"of (3,), got: {cone_axis.shape}"
            )

    # Normalize cone axis:
    cone_axis /= np.linalg.norm(cone_axis)

    if initial_guess is None or initial_guess[2] is None:
        spiral_vector = np.random.random(3)
    else:
        # Check that the input is correct
        try:
            spiral_vector = np.array(initial_guess[2], dtype=float)
        except:
            raise ValueError(
                f"Initial guess for spiral_vector is not array-like: "
                f"{initial_guess[2]}"
            )
        if spiral_vector.shape != (3,):
            raise ValueError(
                f"Initial guess for spiral_vector should have shape "
                f"of (3,), got: {spiral_vector.shape}"
            )

    return [np.concatenate((spin_orientation, [cone_axis]), axis=0), spiral_vector]


################################################################################
#                      Functions that update the argument                      #
#                                                                              #
#   The idea is to call it in the generalized way: x_new = _update(x)          #
################################################################################
def _update_x_ferro(x, search_direction, step=1):
    r"""
    Update the variables for the next step in the ferro case.

    Parameters
    ----------
    x : (I, 3) :numpy:`ndarray`
        Orientation of the spin vectors.
    search_direction : (I*3,) :numpy:`ndarray`
        Search direction computed in L-BFGS step
    step : float
        Size of the step along the search direction.
    """
    x = x.copy()
    a12 = search_direction[::3] * step
    a13 = search_direction[1::3] * step
    a23 = search_direction[2::3] * step

    thetas = np.sqrt(a12**2 + a13**2 + a23**2)

    r = []
    for i in range(len(thetas)):
        theta = thetas[i]

        if theta < np.finfo(float).eps:
            continue

        r = (np.array([-a23[i], a13[i], -a12[i]])) / theta

        x[i] = (
            np.cos(theta) * x[i]
            + np.sin(theta) * np.cross(r, x[i])
            + (1 - np.cos(theta)) * r * (r @ x[i])
        )

    return x


def _update_x_antiferro(x, search_direction, step=1):
    r"""
    Update the variables for the next step in the ferro case.

    Parameters
    ----------
    x : list of two elements: (I+1, 3) :numpy:`ndarray` and (3,) :numpy:`ndarray`
        First element:
            * 0 - I-1:  Orientation of the spin vectors
            * I: cone axis
        Second element: spiral vector
    search_direction : (I*3+6,) :numpy:`ndarray`
        Search direction computed in L-BFGS step
    step : float
        Size of the step along the search direction.
    """
    x_0 = x[0].copy()
    x_1 = x[1]
    a12 = search_direction[:-3:3] * step
    a13 = search_direction[1:-3:3] * step
    a23 = search_direction[2:-3:3] * step

    thetas = np.sqrt(a12**2 + a13**2 + a23**2)

    r = []
    for i in range(len(thetas)):
        theta = thetas[i]

        if theta < np.finfo(float).eps:
            continue

        r = (np.array([-a23[i], a13[i], -a12[i]])) / theta

        x_0[i] = (
            np.cos(theta) * x_0[i]
            + np.sin(theta) * np.cross(r, x_0[i])
            + (1 - np.cos(theta)) * r * (r @ x_0[i])
        )

    return [x_0, x_1]


def _update_x_spiral(x, search_direction, step=1):
    r"""
    Update the variables for the next step in the ferro case.

    Parameters
    ----------
    x : list of two elements: (I+1, 3) :numpy:`ndarray` and (3,) :numpy:`ndarray`
        First element:
            * 0 - I-1:  Orientation of the spin vectors
            * I: cone axis
        Second element: spiral vector
    search_direction : (I*3+6,) :numpy:`ndarray`
        Search direction computed in L-BFGS step
    step : float
        Size of the step along the search direction.
    """
    x_0 = x[0].copy()
    x_1 = x[1].copy()
    a12 = search_direction[:-3:3] * step
    a13 = search_direction[1:-3:3] * step
    a23 = search_direction[2:-3:3] * step

    thetas = np.sqrt(a12**2 + a13**2 + a23**2)

    r = []
    for i in range(len(thetas)):
        theta = thetas[i]

        if theta < np.finfo(float).eps:
            continue

        r = (np.array([-a23[i], a13[i], -a12[i]])) / theta

        x_0[i] = (
            np.cos(theta) * x_0[i]
            + np.sin(theta) * np.cross(r, x_0[i])
            + (1 - np.cos(theta)) * r * (r @ x_0[i])
        )

    x_1 += search_direction[-3:] * step

    return [x_0, x_1]


################################################################################
#                      Functions that unpack the argument                      #
#                                                                              #
# The idea is to call it in the generalized way as                             #
#                                                                              #
#                                  _unpack(x)                                  #
#                                                                              #
# and pass the result to the energy or gradient functions:                     #
#                                                                              #
#                  energy(*_unpack(x)) and _grad(*_unpack(x))                  #
################################################################################
def _unpack_ferro(x):
    r"""
    Unpack the coordinate for the feromagnetic case

    Parameters
    ----------
    x : (I, 3) :numpy:`ndarray`
        Orientation of the spin vectors.

    Returns
    -------
    spin_orientation : (I, 3) :numpy:`ndarray`
        Orientation of the spin vectors.
    """
    return (x,)


def _unpack_antiferro(x):
    r"""
    Unpack parameters for the antiferro case.

    Parameters
    ----------
    x : list of two elements: (I+1, 3) :numpy:`ndarray` and (3,) :numpy:`ndarray`
        First element:
            * 0 - I-1:  Orientation of the spin vectors
            * I: cone axis
        Second element: spiral vector

    Returns
    -------
    spin_orientation : (I, 3) :numpy:`ndarray`
        Orientation of the spin vectors.
    cone_axis : (3,)
        Cone axis.
    spiral_vector : (3,)
        Spiral vector.
    """
    return x[0][:-1], x[0][-1], x[1]


def _unpack_spiral(x):
    r"""
    Unpack parameters for the spiral case.

    Parameters
    ----------
    x : list of two elements: (I+1, 3) :numpy:`ndarray` and (3,) :numpy:`ndarray`
        First element:
            * 0 - I-1:  Orientation of the spin vectors
            * I: cone axis
        Second element: spiral vector

    Returns
    -------
    spin_orientation : (I, 3) :numpy:`ndarray`
        Orientation of the spin vectors.
    cone_axis : (3,)
        Cone axis.
    spiral_vector : (3,)
        Spiral vector.
    """
    return x[0][:-1], x[0][-1], x[1]


################################################################################
#                    Functions that compute torque targets                     #
#                                                                              #
# The idea is to call it in the generalized way as                             #
#                                                                              #
#               torque_targets = _compute_torque_targets(torques)              #
################################################################################


def _compute_torque_target_ferro(torques):
    return (np.linalg.norm(torques, axis=1).max(),)


def _compute_torque_target_antiferro(torques):
    return np.linalg.norm(torques[:-1], axis=1).max(), np.linalg.norm(torques[-1])


def _compute_torque_target_spiral(torques):
    return np.linalg.norm(torques[:-1], axis=1).max(), np.linalg.norm(torques[-1])


################################################################################
#                         Functions that output history                        #
################################################################################


def _default_history_processor_ferro(history, filename=None, nspins=None):
    # Output header
    if history is None:
        if nspins is None:
            _logger.error("If history is None, then nspins is required.")
            raise RunTimeError("If history is None, then nspins is required.")
        header = f"{'Energy':>15} {'Torque':>11}"

        for i in range(nspins):
            header += f" {f'S{i+1}_x':>11} {f'S{i+1}_y':>11} {f'S{i+1}_z':>11}"
        if filename is None:
            print(header, flush=True)
        else:
            with open(filename, "a") as file:
                file.write(header)
    else:
        for i in range(len(history)):
            history[i] = (
                f"{history[i][0]:15.8e} "
                + " ".join([f"{a:11.8f}" for a in history[i][1]])
                + " "
                + " ".join([f"{a:11.8f}" for a in history[i][2].flatten()])
            )
        if filename is None:
            print("\n".join(history), flush=True)
        else:
            with open(filename, "a") as file:
                file.write("\n" + "\n".join(history))


def _default_history_processor_antiferro(history, filename=None, nspins=None):
    # Output header
    if history is None:
        if nspins is None:
            _logger.error("If history is None, then nspins is required.")
            raise RunTimeError("If history is None, then nspins is required.")
        header = f"{'Energy':>15} {'Torque':>11} {'Torque_n':>11}"

        for i in range(nspins):
            header += f" {f'S{i+1}_x':>11} {f'S{i+1}_y':>11} {f'S{i+1}_z':>11}"

        header += f" {'n_x':>11} {'n_y':>11} {'n_z':>11}"
        header += f" {'q_x':>11} {'q_y':>11} {'q_z':>11}"

        if filename is None:
            print(header, flush=True)
        else:
            with open(filename, "a") as file:
                file.write(header)
    else:
        for i in range(len(history)):
            history[i] = (
                # energy
                f"{history[i][0]:15.8e} "
                # torques
                + " ".join([f"{a:11.8f}" for a in history[i][1]])
                + " "
                # I spin vectors and cone axis
                + " ".join([f"{a:11.8f}" for a in history[i][2][0].flatten()])
                + " "
                # spiral vector
                + " ".join([f"{a:11.8f}" for a in history[i][2][1]])
            )
        if filename is None:
            print("\n".join(history), flush=True)
        else:
            with open(filename, "a") as file:
                file.write("\n" + "\n".join(history))


def _default_history_processor_spiral(history, filename=None, nspins=None):
    # Output header
    if history is None:
        if nspins is None:
            _logger.error("If history is None, then nspins is required.")
            raise RunTimeError("If history is None, then nspins is required.")
        header = f"{'Energy':>15} {'Torque':>11} {'Torque_n':>11}"

        for i in range(nspins):
            header += f" {f'S{i+1}_x':>11} {f'S{i+1}_y':>11} {f'S{i+1}_z':>11}"

        header += f" {'n_x':>11} {'n_y':>11} {'n_z':>11}"
        header += f" {'q_x':>11} {'q_y':>11} {'q_z':>11}"

        if filename is None:
            print(header, flush=True)
        else:
            with open(filename, "a") as file:
                file.write(header)
    else:
        for i in range(len(history)):
            history[i] = (
                # energy
                f"{history[i][0]:15.8e} "
                # torques
                + " ".join([f"{a:11.8f}" for a in history[i][1]])
                + " "
                # I spin vectors and cone axis
                + " ".join([f"{a:11.8f}" for a in history[i][2][0].flatten()])
                + " "
                # spiral vector
                + " ".join([f"{a:11.8f}" for a in history[i][2][1]])
            )
        if filename is None:
            print("\n".join(history), flush=True)
        else:
            with open(filename, "a") as file:
                file.write("\n" + "\n".join(history))


################################################################################
#                                 Energy class                                 #
#                                                                              #
# The idea is to create it from SpinHamiltonian as                             #
#                                                                              #
#                           energy = Energy(spinham)                           #
#                                                                              #
# and then have the routines for the energy optimization with the call         #
# signatures                                                                   #
#                                                                              #
#                energy.optimize(initial_guess, *args, **kwargs)               #
################################################################################
class Energy:
    r"""
    Energy of the :py:class:`.SpinHamiltonian`.

    Includes methods for energy optimization for three main types of the ground state
    and for the computation of the energy with the given spin arrangements.

    Being created from a given :py:class:`.SpinHamiltonian` it is completely independent
    from it. i.e. changes in the instance of the :py:class:`.SpinHamiltonian` class do
    not affect the instance of the :py:class:`.Energy` class.
    """

    def __init__(self, spinham: SpinHamiltonian):
        # Parameters of the Hamiltonian, private
        self._factor = spinham.exchange_factor
        self._atom_indices = dict(
            [(atom, i) for i, atom in enumerate(spinham.magnetic_atoms)]
        )

        # Parameters of the Hamiltonian, exposed to public as properties
        self._g_factors = [atom.g_factor for atom in spinham.magnetic_atoms]
        self._spins = [atom.spin for atom in spinham.magnetic_atoms]

        # Force double counting notation
        previous_dc = spinham.double_counting
        spinham.double_counting = True

        # Get all bonds of the Hamiltonian.
        self._bonds = []
        for atom1, atom2, R, J in spinham.exchange_like:
            i = self._atom_indices[atom1]
            j = self._atom_indices[atom2]
            d = R @ spinham.cell
            self._bonds.append([i, j, d, J])

        # Return original noatation of SpinHamiltonian
        spinham.double_counting = previous_dc

        # Minimisation settings, exposed to public as properties
        self._m = 10

        # Parameters for Wolfe conditions
        self._wolfe_iter_max = 10000
        self._wolfe_c1 = 1e-4
        self._wolfe_c2 = 0.9

        # Minimization settings, private
        # Gradient size depends on the expected ground state.
        self._gradient_size = None  # Have to be initialized at the beginning of every optimization function
        self._torques_size = None  # Have to be initialized at the beginning of every optimization function

        # External physical conditions
        self._magnetic_field = None

        # To be sorted
        self._d_vec = None  # Have to be initialized at the beginning of every optimization function
        self._y_vec = None  # Have to be initialized at the beginning of every optimization function
        self._rho = np.zeros((self._m), dtype=float)
        self._gamma = np.zeros((self._m), dtype=float)

    ################################################################################
    #                             Minimization settings                            #
    ################################################################################

    @property
    def m(self) -> int:
        r"""
        Amount of steps to keep in memory for the estimation of the Hessian matrix in
        L-BFGS algorithm.

        Returns
        -------
        m : int
        """

        return self._m

    @m.setter
    def m(self, new_value):
        try:
            new_m = int(new_value)
        except:
            raise ValueError(
                "New size of memory storage for L-BFGS is not "
                f"an integer:{new_value}."
            )
        self._m = new_m

    @property
    def wolfe_c1(self) -> float:
        r"""
        Parameter for the Wolfe conditions:

        .. math::
          f(\boldsymbol{x} + \lambda\boldsymbol{p})
          \le
          f(\boldsymbol{x})
          +
          c_1  \lambda \nabla_{\boldsymbol{x}}f(\boldsymbol{x}) \cdot \boldsymbol{p}

        Default value is :math:`10^{-4}`

        Returns
        -------
        c1 : float
        """

        return self._wolfe_c1

    @wolfe_c1.setter
    def wolfe_c1(self, new_value):
        try:
            new_c1 = float(new_value)
        except:
            raise ValueError(
                "New c1 coefficient for Wolfe conditions is not "
                f"a float:{new_value}."
            )
        self._wolfe_c1 = new_c1

    @property
    def wolfe_c2(self) -> float:
        r"""
        Parameter for the Wolfe conditions:


        .. math::
          -\nabla_{\boldsymbol{x}}f(\boldsymbol{x} + \lambda\boldsymbol{p}) \cdot \boldsymbol{p}
          \le
          -c_2
          \nabla_{\boldsymbol{x}}f(\boldsymbol{x}) \cdot \boldsymbol{p}

        Default value is :math:`0.9`

        Returns
        -------
        c2 : float
        """

        return self._wolfe_c1

    @wolfe_c2.setter
    def wolfe_c2(self, new_value):
        try:
            new_c2 = float(new_value)
        except:
            raise ValueError(
                "New c2 coefficient for Wolfe conditions is not "
                f"a float:{new_value}."
            )
        self._wolfe_c2 = new_c2

    ################################################################################
    #                         External physical conditions                         #
    ################################################################################

    @property
    def magnetic_field(self):
        r"""
        External magnetic field. Given in the units of Tesla.

        Returns
        =======
        magnetic_field : (3,) :numpy:`ndarray` or ``None``
            magnetic_field :math:`\boldsymbol{H}`.
        """
        return self._magnetic_field

    @magnetic_field.setter
    def magnetic_field(self, new_value):
        if new_value is None:
            self._magnetic_field = None
        else:
            try:
                new_field = np.array(new_value, dtype=float)
            except:
                raise ValueError(f"New magnetic field is not array-like: {new_value}")

            if new_field.shape != (3,):
                raise ValueError(
                    "New magnetic field has to be a 3 component vector, "
                    f"got {new_field.shape}"
                )

            self._magnetic_field = new_field

    ################################################################################
    #                         Parameters of the Hamiltonian                        #
    ################################################################################

    @property
    def g_factors(self):
        r"""
        g factor for all magnetic atoms.

        Returns
        -------
        g_factors : (I,) :numpy:`ndarray`
            Order is the same as in :py:attr:`.Energy.spins`.
        """
        return self._g_factors

    @g_factors.setter
    def g_factors(self, new_value):
        if not isinstance(Iterable):
            raise ValueError(f"New g factors must be an Iterable: {new_value}")

        if not len(new_value) == len(self._spins):
            raise ValueError(
                f"Expecten {len(self._spins)} g factors, got {len(new_value)}"
            )

        for i, g_factor in enumerate(new_value):
            try:
                new_value[i] = int(g_factor)
            except:
                raise ValueError(
                    f"One of the new g_factors (index {i}) is not "
                    f"an integer:{g_factor}."
                )
        self._g_factors = new_value

    @property
    def spins(self):
        r"""
        Value of spin for all magnetic atoms.

        Returns
        -------
        spins : (I,) :numpy:`ndarray`
            Order is the same as in :py:attr:`.SpinHamiltonian.magnetic_atoms`.
        """
        return self._spins

    @spins.setter
    def spins(self, new_value):
        if not isinstance(Iterable):
            raise ValueError(f"New spins must be an Iterable: {new_value}")

        if not len(new_value) == len(self._spins):
            raise ValueError(f"Expected {len(self._spins)} spins, got {len(new_value)}")

        for i, spin in enumerate(new_value):
            try:
                new_value[i] = int(spin)
            except:
                raise ValueError(
                    f"One of the new spins (index {i}) is not " f"an integer:{spin}."
                )
        self._spins = new_value

    ################################################################################
    #                              Gradients of energy                             #
    # The idea is to have functions that have a call signatures the same as        #
    # corresponding energy functions with two additional keyword arguments:        #
    # "gradient_vector" and "torques". If both gradient_vector and torques are     #
    # given, then the update is done in place and the function return nothing.     #
    # Otherwise it returns gradient_vector and torques, leaving the given part     #
    # intact.                                                                      #
    ################################################################################

    def _ferro_grad(self, spin_orientation, gradient_vector=None, torques=None):
        r"""
        Gradient of the ferro case looks like:

        [ t_0z, -t_0y, t_0x, ..., t_(I-1)z, -t_(I-1)y ]

        where t_ix, t_iy, t_iz - are components of torque for ith spin in the
        unit cell (i = 0, I-1).

        torques have the form:

        [
            [t_0x, t_0y, t_0z],
            ...,
            [t_(I-1)x, t_(I-1)y, t_(I-1)z]
        ]
        """
        if gradient_vector is None or torques is None:
            gradient_vector = np.zeros((self._gradient_size), dtype=float)
            torques = np.zeros((len(self._spins), 3), dtype=float)
            return_results = True
        else:
            return_results = False

        for i in range(len(spin_orientation)):
            h = 1e-4
            gradient = []
            for j in range(3):
                spin_orientation[i][j] -= h
                f_minus_h = self.ferro(spin_orientation)
                spin_orientation[i][j] += 2 * h
                f_plus_h = self.ferro(spin_orientation)
                gradient.append((f_plus_h - f_minus_h) / 2 / h)
                spin_orientation[i][j] -= h

            t = np.cross(spin_orientation[i], gradient)
            for j in range(0, 3):
                torques[i][j] = t[j]
            gradient_vector[3 * i] = t[2]
            gradient_vector[3 * i + 1] = -t[1]
            gradient_vector[3 * i + 2] = t[0]

        if return_results:
            return gradient_vector, torques

    def _antiferro_grad(
        self,
        spin_orientation,
        cone_axis,
        spiral_vector,
        gradient_vector=None,
        torques=None,
    ):
        r"""
        Gradient of the antiferro case looks like:

        [ t_0z, -t_0y, t_0x, ..., t_(I-1)z, -t_(I-1)y,
          t_(I-1)x, t_nz, -t_ny, t_nx]

        where t_ix, t_iy, t_iz - are components of torque for ith spin in the
        unit cell (i = 0, I-1); t_nx, t_ny, t_nz 0 - are the components of the "torque"
        on the cone axis.

        torques have the form:

        [
            [t_0x, t_0y, t_0z],
            ...,
            [t_(I-1)x, t_(I-1)y, t_(I-1)z],
            [t_nx, t_ny, t_nz]
        ]
        """

        if gradient_vector is None or torques is None:
            gradient_vector = np.zeros((self._gradient_size), dtype=float)
            torques = np.zeros((self._torque_size, 3), dtype=float)
            return_results = True
        else:
            return_results = False

        # Compute gradient for spin_orientation
        for i in range(len(spin_orientation)):
            h = 1e-4
            gradient = []
            for j in range(3):
                spin_orientation[i][j] -= h
                f_minus_h = self.antiferro(spin_orientation, cone_axis, spiral_vector)
                spin_orientation[i][j] += 2 * h
                f_plus_h = self.antiferro(spin_orientation, cone_axis, spiral_vector)
                gradient.append((f_plus_h - f_minus_h) / 2 / h)
                spin_orientation[i][j] -= h

            t = np.cross(spin_orientation[i], gradient)
            for j in range(0, 3):
                torques[i][j] = t[j]
            gradient_vector[3 * i] = t[2]
            gradient_vector[3 * i + 1] = -t[1]
            gradient_vector[3 * i + 2] = t[0]

        # Compute gradient for cone_axis
        h = 1e-4
        gradient = []
        for j in range(3):
            cone_axis[j] -= h
            f_minus_h = self.antiferro(spin_orientation, cone_axis, spiral_vector)
            cone_axis[j] += 2 * h
            f_plus_h = self.antiferro(spin_orientation, cone_axis, spiral_vector)
            gradient.append((f_plus_h - f_minus_h) / 2 / h)
            cone_axis[j] -= h

        t = np.cross(cone_axis, gradient)
        for j in range(0, 3):
            torques[-1][j] = t[j]
        gradient_vector[-6] = t[2]
        gradient_vector[-5] = -t[1]
        gradient_vector[-4] = t[0]

        if return_results:
            return gradient_vector, torques

    def _spiral_grad(
        self,
        spin_orientation,
        cone_axis,
        spiral_vector,
        gradient_vector=None,
        torques=None,
    ):
        r"""
        Gradient of the spiral case looks like:

        [ t_0z, -t_0y, t_0x, ..., t_(I-1)z, -t_(I-1)y,
          t_(I-1)x, t_nz, -t_ny, t_nx,
          dE/dq_x, dE/dq_y, dE/dq_z ]

        where t_ix, t_iy, t_iz - are components of torque for ith spin in the
        unit cell (i = 0, I-1); t_nx, t_ny, t_nz 0 - are the components of the "torque"
        on the cone axis and last three elements are just partial derivatives
        with respect to the components of the spiral vector (computed analytically).

        torques have the form:

        [
            [t_0x, t_0y, t_0z],
            ...,
            [t_(I-1)x, t_(I-1)y, t_(I-1)z],
            [t_nx, t_ny, t_nz]
        ]
        """

        if gradient_vector is None or torques is None:
            gradient_vector = np.zeros((self._gradient_size), dtype=float)
            torques = np.zeros((self._torque_size, 3), dtype=float)
            return_results = True
        else:
            return_results = False

        # Compute gradient for spin_orientation
        for i in range(len(spin_orientation)):
            h = 1e-4
            gradient = []
            for j in range(3):
                spin_orientation[i][j] -= h
                f_minus_h = self.spiral(spin_orientation, cone_axis, spiral_vector)
                spin_orientation[i][j] += 2 * h
                f_plus_h = self.spiral(spin_orientation, cone_axis, spiral_vector)
                gradient.append((f_plus_h - f_minus_h) / 2 / h)
                spin_orientation[i][j] -= h

            t = np.cross(spin_orientation[i], gradient)
            for j in range(0, 3):
                torques[i][j] = t[j]
            gradient_vector[3 * i] = t[2]
            gradient_vector[3 * i + 1] = -t[1]
            gradient_vector[3 * i + 2] = t[0]

        # Compute gradient for cone_axis
        h = 1e-4
        gradient = []
        for j in range(3):
            cone_axis[j] -= h
            f_minus_h = self.spiral(spin_orientation, cone_axis, spiral_vector)
            cone_axis[j] += 2 * h
            f_plus_h = self.spiral(spin_orientation, cone_axis, spiral_vector)
            gradient.append((f_plus_h - f_minus_h) / 2 / h)
            cone_axis[j] -= h

        t = np.cross(cone_axis, gradient)
        for j in range(0, 3):
            torques[-1][j] = t[j]
        gradient_vector[-6] = t[2]
        gradient_vector[-5] = -t[1]
        gradient_vector[-4] = t[0]

        # Compute gradient for spiral_vector
        K = np.array(
            [
                [0, -cone_axis[2], cone_axis[1]],
                [cone_axis[2], 0, -cone_axis[0]],
                [-cone_axis[1], cone_axis[0], 0],
            ]
        )
        K_sq = K @ K

        gradient = np.zeros(3, dtype=float)

        for i, j, d, J in self._bonds:
            C = -np.cos(spiral_vector @ d) / 2 * (
                K @ J.matrix @ K_sq + K_sq @ J.matrix @ K
            ) + np.sin(spiral_vector @ d) / 2 * (
                K @ J.matrix @ K - K_sq @ J.matrix @ K_sq
            )
            gradient += d * (
                self._factor
                * self._spins[i]
                * self._spins[j]
                * np.einsum("i,ij,j", spin_orientation[i], C, spin_orientation[j])
            )
        # print(f"\n\n{gradient}\n\n")

        gradient_vector[-3:] = gradient

        if return_results:
            return gradient_vector, torques

    ################################################################################
    #                                    Energy                                    #
    ################################################################################

    def ferro(self, spin_orientation):
        r"""
        Computes energy of the generalized ferromagnetic state for given
        ``spin_orientation``.

        Parameters
        ----------
        spin_orientation : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only direction of the vectors
            is important.

        Returns
        -------
        energy :  float.
            Energy of the Hamiltonian for the given ``spin_orientation``
        """

        # Check that the input is correct
        try:
            spin_orientation = np.array(spin_orientation, dtype=float)
        except:
            raise ValueError(f"spin_orientation is not array-like: {spin_orientation}")

        if spin_orientation.shape[-1] != 3:
            raise ValueError(
                f"spin_orientation has to have the last dimension of "
                f"size 3, got {spin_orientation.shape[-1]}"
            )

        # Make the input array have (I,3) shape for I = 1
        if len(self._spins) == 1 and spin_orientation.shape == (3,):
            spin_orientation = np.array([spin_orientation])

        if len(self._spins) != spin_orientation.shape[0]:
            raise ValueError(
                f"Expected {len(self._spins)} spin orientations, "
                f"got {spin_orientation.shape[0]}"
            )

        for i in range(len(spin_orientation)):
            spin_orientation[i] /= np.linalg.norm(spin_orientation[i])

        energy = 0

        # Compute exchange and single-ion-like anisotropy energy
        for i, j, _, J in self._bonds:
            energy += (
                self._factor
                * self._spins[i]
                * self._spins[j]
                * np.einsum(
                    "i,ij,j", spin_orientation[i], J.matrix, spin_orientation[j]
                )
            )

        # Compute zeeman energy
        if self.magnetic_field is not None:
            energy += BOHR_MAGNETON * np.einsum(
                "i,j,ij", self._g_factors, self.magnetic_field, spin_orientation
            )

        return energy

    def antiferro(self, spin_orientation, cone_axis, spiral_vector):
        r"""
        Computes energy of the antiferro cone state for given
        ``spin_orientation``, ``cone_axis`` and ``spiral_vector``.

        Parameters
        ----------
        spin_orientation : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only direction of the vectors
            is important.
        cone_axis : (3,) |array-like|_
            Orientation of the cone axis, only the direction is important.
        spiral_vector : (3,) |array-like|_
            Spiral vector.

        Returns
        -------
        energy :  float.
            Energy of the Hamiltonian for the given ``spin_orientation``, ``cone_axis``
            and ``spiral_vector``.

        Notes
        -----
        Spiral vector is restricted to :math:`2\boldsymbol{q}=\boldsymbol{G}`.
        """

        # Check that the input is correct (spins)
        try:
            spin_orientation = np.array(spin_orientation, dtype=float)
        except:
            raise ValueError(f"spin_orientation is not array-like: {spin_orientation}")

        if spin_orientation.shape[-1] != 3:
            raise ValueError(
                f"spin_orientation has to have the last dimension of "
                f"size 3, got {spin_orientation.shape[-1]}"
            )

        # Check that the input is correct (cone_axis)
        try:
            cone_axis = np.array(cone_axis, dtype=float)
        except:
            raise ValueError(f"cone_axis is not array-like: {cone_axis}")

        if cone_axis.shape != (3,):
            raise ValueError(
                f"cone_axis must have the shape of (3,) " f"got {cone_axis.shape}"
            )

        # Check that the input is correct (spiral vector)
        try:
            spiral_vector = np.array(spiral_vector, dtype=float)
        except:
            raise ValueError(f"spiral_vector is not array-like: {spiral_vector}")

        if spiral_vector.shape != (3,):
            raise ValueError(
                f"spiral_vector must have the shape of (3,) "
                f"got {spiral_vector.shape}"
            )

        # Make the input array have (I,3) shape for I = 1
        if len(self._spins) == 1 and spin_orientation.shape == (3,):
            spin_orientation = np.array([spin_orientation])

        if len(self._spins) != spin_orientation.shape[0]:
            raise ValueError(
                f"Expected {len(self._spins)} spin orientations, "
                f"got {spin_orientation.shape[0]}"
            )

        # Normalize spin orientations
        for i in range(len(spin_orientation)):
            spin_orientation[i] /= np.linalg.norm(spin_orientation[i])

        # Normalize cone axis
        cone_axis /= np.linalg.norm(cone_axis)

        K = np.array(
            [
                [0, -cone_axis[2], cone_axis[1]],
                [cone_axis[2], 0, -cone_axis[0]],
                [-cone_axis[1], cone_axis[0], 0],
            ]
        )
        K_sq = K @ K
        IK_sq = np.eye(3, dtype=float) + K_sq

        energy = 0

        # Compute exchange and single-ion-like anisotropy energy
        for i, j, d, J in self._bonds:
            C_dij = (
                IK_sq @ J.matrix @ IK_sq
                + K_sq @ J.matrix @ K_sq * np.cos(spiral_vector @ d)
                - K_sq @ J.matrix @ K * np.sin(spiral_vector @ d)
            )

            energy += (
                self._factor
                * self._spins[i]
                * self._spins[j]
                * np.einsum("i,ij,j", spin_orientation[i], C_dij, spin_orientation[j])
            )

        # Compute zeeman energy
        if self.magnetic_field is not None:
            so_prime = np.einsum(
                "jk,ik ->ij", (np.eye(3, dtype=float) + IK_sq), spin_orientation
            )
            energy += BOHR_MAGNETON * np.einsum(
                "i,j,ij", self._g_factors, self.magnetic_field, so_prime
            )

        return energy

    def spiral(self, spin_orientation, cone_axis, spiral_vector):
        r"""
        Computes energy of the spiral cone state for given
        ``spin_orientation``, ``cone_axis`` and ``spiral_vector``.

        Parameters
        ----------
        spin_orientation : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only direction of the vectors
            is important.
        cone_axis : (3,) |array-like|_
            Orientation of the cone axis, only the direction is important.
        spiral_vector : (3,) |array-like|_
            Spiral vector.

        Returns
        -------
        energy :  float.
            Energy of the Hamiltonian for the given ``spin_orientation``, ``cone_axis``
            and ``spiral_vector``.
        """

        # Check that the input is correct (spins)
        try:
            spin_orientation = np.array(spin_orientation, dtype=float)
        except:
            raise ValueError(f"spin_orientation is not array-like: {spin_orientation}")

        if spin_orientation.shape[-1] != 3:
            raise ValueError(
                f"spin_orientation has to have the last dimension of "
                f"size 3, got {spin_orientation.shape[-1]}"
            )

        # Check that the input is correct (cone_axis)
        try:
            cone_axis = np.array(cone_axis, dtype=float)
        except:
            raise ValueError(f"cone_axis is not array-like: {cone_axis}")

        if cone_axis.shape != (3,):
            raise ValueError(
                f"cone_axis must have the shape of (3,) " f"got {cone_axis.shape}"
            )

        # Check that the input is correct (spiral vector)
        try:
            spiral_vector = np.array(spiral_vector, dtype=float)
        except:
            raise ValueError(f"spiral_vector is not array-like: {spiral_vector}")

        if spiral_vector.shape != (3,):
            raise ValueError(
                f"spiral_vector must have the shape of (3,) "
                f"got {spiral_vector.shape}"
            )

        # Make the input array have (I,3) shape for I = 1
        if len(self._spins) == 1 and spin_orientation.shape == (3,):
            spin_orientation = np.array([spin_orientation])

        if len(self._spins) != spin_orientation.shape[0]:
            raise ValueError(
                f"Expected {len(self._spins)} spin orientations, "
                f"got {spin_orientation.shape[0]}"
            )

        # Normalize spin orientations
        for i in range(len(spin_orientation)):
            spin_orientation[i] /= np.linalg.norm(spin_orientation[i])

        # Normalize cone axis
        cone_axis /= np.linalg.norm(cone_axis)

        K = np.array(
            [
                [0, -cone_axis[2], cone_axis[1]],
                [cone_axis[2], 0, -cone_axis[0]],
                [-cone_axis[1], cone_axis[0], 0],
            ]
        )
        K_sq = K @ K
        IK_sq = np.eye(3, dtype=float) + K_sq

        energy = 0

        # Compute exchange and single-ion-like anisotropy energy
        for i, j, d, J in self._bonds:
            # print(
            #     spiral_vector @ d, np.sin(spiral_vector @ d), np.cos(spiral_vector @ d)
            # )
            C_dij = (
                IK_sq @ J.matrix @ IK_sq
                - (K @ J.matrix @ K - K_sq @ J.matrix @ K_sq)
                / 2
                * np.cos(spiral_vector @ d)
                - (K @ J.matrix @ K_sq + K_sq @ J.matrix @ K)
                / 2
                * np.sin(spiral_vector @ d)
            )

            energy += (
                self._factor
                * self._spins[i]
                * self._spins[j]
                * np.einsum("i,ij,j", spin_orientation[i], C_dij, spin_orientation[j])
            )

        # Compute zeeman energy
        if self.magnetic_field is not None:
            so_prime = np.einsum(
                "jk,ik ->ij", (np.eye(3, dtype=float) + IK_sq), spin_orientation
            )
            energy += BOHR_MAGNETON * np.einsum(
                "i,j,ij", self._g_factors, self.magnetic_field, so_prime
            )

        return energy

    ################################################################################
    #                             Optimization routines                            #
    ################################################################################

    def optimize(
        self,
        case,
        initial_guess=None,
        tolerance=None,
        max_iterations=None,
        save_history=False,
        history_filename=None,
        history_step=1000,
        antiferro_q=None,
    ):
        r"""
        Find the minima of the energy assuming the type of the ground state based on
        ``case`` argument.

        Parameters
        ----------
        case : str or int {"ferro", 0, "antiferro", 1, "spiral", 2}
            Three cases are supported:

            * Generalized ferromagnetic: ``"ferro"`` or ``0``
            * Generalized antiferromagnetic: ``"antiferro"`` or ``1``
            * Spiral cone: ``"spiral"`` or ``2``
        initial_guess : optional
            Initial guess for the relevant variables. Expected format depends on the
            ``case``:

            * ``case="ferro"`` spin_orientation

              spin_orientation : (I, 3) or (3,) |array-like|_ or None
                  Initial guess for the orientation of spin vectors.
                  If none provided, then random orientation is chosen.
                  The vectors are normalized to one, i.e. only direction of the vectors
                  is important.
                  If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are
                  accepted.

            * ``case ="antiferro"`` tuple of (spin_orientation, cone_axis)

              spin_orientation : (I, 3) or (3,) |array-like|_ or None
                  Initial guess for the orientation of spin vectors.
                  If none provided, then random orientation is chosen.
                  The vectors are normalized to one, i.e. only direction of the vectors
                  is important.
                  If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are
                  accepted.
              cone_axis : (3,) |array-like|_
                  Cone axis. Only direction is important.

            * ``case ="spiral"`` tuple of (spin_orientation, cone_axis, spiral_vector)

              spin_orientation : (I, 3) or (3,) |array-like|_ or None
                  Initial guess for the orientation of spin vectors.
                  If none provided, then random orientation is chosen.
                  The vectors are normalized to one, i.e. only direction of the vectors
                  is important.
                  If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are
                  accepted.
              cone_axis : (3,) |array-like|_
                  Cone axis. Only direction is important.
              spiral_vector : (3,) |array-like|_
                  Spiral vector :math:`\boldsymbol{q}`. Both orientation and length are
                  important.

        tolerance : tuple of floats
            Tolerance parameters.

            * ``case="ferro"`` tuple of (tol_energy_diff, tol_torque_max)

              tol_energy_diff : float, default 1e-5
                  tolerance for the difference between the
                  energies of two consecutive minimization steps
              tol_torque_max : float, default 1e-5
                  Tolerance for the maximum torque among the spins of the unit cell.

            * ``case="antiferro"``
              FIXME

            * ``case="ferro"`` tuple of (tol_energy_diff, tol_torque_max, tol_torque_cone)

              tol_energy_diff : float, default 1e-5
                  tolerance for the difference between the
                  energies of two consecutive minimization steps
              tol_torque_max : float, default 1e-5
                  Tolerance for the maximum torque among the spins of the unit cell.
              tol_torque_cone : float, default 1e-5
                  Tolerance for the torque on the cone axis.

        max_iterations : int, optional
            Maximum amount of iterations for the minimization algorithm.
            By default the algorithm will run until the convergence is reached.
        save_history : bool, default False
            Whether to output the steps of the minimization algorithm.
        history_filename : str, optional
            Filename for saving history of the minimization. By default history is
            directed to ``sys.stdout`` via ``print()`` function.
        history_step : int, default 100
            The history is printed every ``history_step`` steps.

        antiferro_q : (3,) |array-like|_
            One of the 26 possible q vectors of the sntiferromagnetic case.


        Returns
        -------
        spin_orientation : (I, 3) :numpy:`ndarray`
            Spin orientation that corresponds to the found local minima of energy
            assuming generalized ferromagnetic state. Returned in every case.
        cone_axis : (3,) :numpy:`ndarray`
            Cone axis that correspond to the local minima of energy. Returned if
            ``case = "antiferro"`` or ``case = "spiral"``
        spiral_vector : (3,)  :numpy:`ndarray`
            Spiral vector that correspond to the local minima of energy. Returned if
            ``case = "spiral"``
        """

        # Check the correctness of the initial guess if any.
        # Make a random one otherwise.

        if case in [0, "ferro"]:
            x = _starting_ferro(len(self._spins), initial_guess)

            self._gradient_size = 3 * len(self._spins)
            self._torque_size = len(self._spins)

            energy = self.ferro
            _grad = self._ferro_grad
            _unpack = _unpack_ferro
            _update = _update_x_ferro
            _compute_torque_target = _compute_torque_target_ferro

            # Switch to the default history processor if necessary
            if save_history:
                history_processor = _default_history_processor_ferro

            if tolerance is None:
                tolerance = (1e-5, 1e-5)
            elif len(tolerance) != 2:
                raise ValueError(
                    f"Expected tuple of two floats for tolerance, got {tolerance}"
                )
        elif case in [1, "antiferro"]:
            x = _starting_antiferro(len(self._spins), antiferro_q, initial_guess)

            self._gradient_size = 3 * len(self._spins) + 3
            self._torque_size = len(self._spins) + 1

            energy = self.antiferro
            _grad = self._antiferro_grad
            _unpack = _unpack_antiferro
            _update = _update_x_antiferro
            _compute_torque_target = _compute_torque_target_antiferro

            # Switch to the default history processor if necessary
            if save_history:
                history_processor = _default_history_processor_antiferro
            if tolerance is None:
                tolerance = (1e-5, 1e-5, 1e-5)
            elif len(tolerance) != 3:
                raise ValueError(
                    f"Expected tuple of three floats for tolerance, got {tolerance}"
                )
        elif case in [2, "spiral"]:
            x = _starting_spiral(len(self._spins), initial_guess)

            self._gradient_size = 3 * len(self._spins) + 6
            self._torque_size = len(self._spins) + 1

            energy = self.spiral
            _grad = self._spiral_grad
            _unpack = _unpack_spiral
            _update = _update_x_spiral
            _compute_torque_target = _compute_torque_target_spiral

            # Switch to the default history processor if necessary
            if save_history:
                history_processor = _default_history_processor_spiral
            if tolerance is None:
                tolerance = (1e-5, 1e-5, 1e-5)
            elif len(tolerance) != 3:
                raise ValueError(
                    f"Expected tuple of three floats for tolerance, got {tolerance}"
                )
        else:
            raise ValueError(
                'case must be either "ferro" or "antiferro" or "spiral" or '
                f"0 or 1 or 2. Got {case}"
            )

        # Print header for the history
        if save_history:
            history_processor(None, nspins=len(self._spins), filename=history_filename)

        # Set the iterators
        # Note: iteration counter and k are distinct: k might be set to zero during
        # the algorithm's run.
        k = 0
        iteration_counter = 0

        # Set initial tracking variables
        step = None
        previous_gradient_vector = None
        previous_search_direction = None
        energy_prev = energy(*_unpack(x))
        self._d_vec = np.zeros((self._m, self._gradient_size), dtype=float)
        self._y_vec = np.zeros((self._m, self._gradient_size), dtype=float)
        torques = np.zeros((self._torque_size, 3), dtype=float)
        gradient_vector = np.zeros((self._gradient_size), dtype=float)
        search_direction = np.zeros((self._gradient_size), dtype=float)

        # Set initial convergence targets for torques so the loop will be started.
        # For the energy the track is done separately
        torque_targets = [2 * a for a in tolerance[1:]]
        energy_diff = tolerance[0] * 2

        # Compute condition for the loop
        condition = energy_diff > tolerance[0]
        for i in range(len(torque_targets)):
            condition = condition or torque_targets[i] > tolerance[i + 1]

        # Save initial guess and its energy, conv_targets are meaningless
        # but saved so the history will have the same format for each step
        if save_history:
            history = [[energy_prev, torque_targets, x]]

        while condition:
            # Check whether number of iterations is exceeded
            if max_iterations is not None and iteration_counter >= max_iterations:
                _logger.warning(
                    f"Maximum iteration is reached. "
                    f"Torques: {torque_targets}, "
                    f"Energy difference: {energy_diff:.5e}."
                )
                return _unpack(x)

            # Compute gradient and torques
            _grad(*_unpack(x), gradient_vector=gradient_vector, torques=torques)

            # Compute search direction
            k, search_direction = self._compute_search_direction(
                k,
                gradient_vector=gradient_vector,
                previous_lambda=step,
                previous_gradient_vector=previous_gradient_vector,
                previous_search_direction=previous_search_direction,
            )

            # Compute size of the step along the search direction
            step = self._lambda(
                x,
                search_direction,
                energy=energy,
                _unpack=_unpack,
                _update=_update,
                _grad=_grad,
            )

            # Update the generalized coordinate
            x = _update(x, search_direction, step)

            # Compute convergence targets
            energy_current = energy(*_unpack(x))
            energy_diff = abs(energy_prev - energy_current)
            torque_targets = _compute_torque_target(torques)

            # Compute condition for the loop
            condition = energy_diff > tolerance[0]
            for i in range(len(torque_targets)):
                condition = condition or torque_targets[i] > tolerance[i + 1]

            # Keep values for the next loop
            previous_gradient_vector = gradient_vector.copy()
            previous_search_direction = search_direction.copy()
            energy_prev = energy_current

            # Update iterators
            k += 1
            iteration_counter += 1

            if save_history:
                # Update history
                history.append(
                    [
                        energy_current,
                        torque_targets,
                        x,
                    ]
                )
                # Output history is required
                if iteration_counter % history_step == 0 and iteration_counter > 0:
                    history_processor(history, filename=history_filename)
                    history = []

        # Output remaining part of history
        if save_history:
            history_processor(history, filename=history_filename)

        # Return minimized configuration
        return _unpack(x)

    ################################################################################
    #                                 L-BFGS update                                #
    ################################################################################
    def _compute_search_direction(
        self,
        k,
        gradient_vector,
        previous_lambda=None,
        previous_gradient_vector=None,
        previous_search_direction=None,
    ):
        r"""
        Compute search direction with L-BFGS algorithm. The details of the
        implementation are taken from [1]_.
        Universal and does not depend on the choice of the ground state type.

        Parameters
        ----------
        k : int
            Index of the minimization cycle.
        gradient_vector : (M,) :numpy:`ndarray`
            Gradient vector.
        previous_gradient_vector : (M,) :numpy:`ndarray`
            Gradient vector for k-1.
        previous_search_direction : (M,) :numpy:`ndarray`
            Search direction for k-1.

        Returns
        -------
        k : int
            Index of the minimization cycle (can be set to 0 at some of the steps).
        search_direction : (M,)
            Search direction computed for the step k.

        References
        ----------
        .. [1] Ivanov, A.V., Uzdin, V.M. and Jónsson, H., 2021.
            Fast and robust algorithm for energy minimization of spin systems applied
            in an analysis of high temperature spin configurations in terms of skyrmion
            density.
            Computer Physics Communications, 260, p.107749.
        """

        if k != 0 and (
            previous_gradient_vector is None
            or previous_search_direction is None
            or previous_lambda is None
        ):
            raise ValueError("Need info about previous gradient and search direction")

        n = k % self._m

        if k == 0:
            self._d_vec = np.zeros((self._m, self._gradient_size), dtype=float)
            self._y_vec = np.zeros((self._m, self._gradient_size), dtype=float)
            self._rho = np.zeros((self._m), dtype=float)
            self._gamma = np.zeros((self._m), dtype=float)
            search_direction = gradient_vector
        else:
            self._d_vec[n] = previous_lambda * previous_search_direction
            self._y_vec[n] = gradient_vector - previous_gradient_vector
            self._rho[n] = 1 / (self._d_vec[n] @ self._y_vec[n])

            if self._rho[n] < 0:
                return (
                    0,
                    self._compute_search_direction(0, gradient_vector=gradient_vector)[
                        1
                    ],
                )

            q = gradient_vector.copy()

            for l in range(self._m - 1, -1):
                j = (l + n + 1) % self._m
                self._gamma[j] = self._rho[j] * (self._d_vec[j] @ q)
                q = q - self._gamma[j] * self._y_vec[j]
            search_direction = q / self._rho[n] / (np.linalg.norm(self._y_vec[n]))
            for l in range(0, self._m):
                if k < self._m:
                    j = l
                else:
                    j = (l + n + 1) % self._m
                search_direction = search_direction + self._d_vec[j] * (
                    self._gamma[j] - self._rho[j] * (self._y_vec[j] @ search_direction)
                )

        return k, search_direction

    ################################################################################
    #                   Inexact line search with Wolfe conditions                  #
    ################################################################################
    def _lambda(self, x, search_direction, energy, _unpack, _update, _grad):
        r"""
        Compute the step length across given search direction. Universal and does not
        depend on the choice of the ground state type.
        """
        step = 1
        fx = energy(*_unpack(x))
        x_new = _update(x, search_direction, step)
        nabla_new = _grad(*_unpack(x_new))[0]
        nabla = _grad(*_unpack(x))[0]
        i = 0
        tmp = [[], [], []]
        while energy(*_unpack(x_new)) > fx + (
            self._wolfe_c1 * step * nabla.T @ search_direction
        ) or abs(nabla_new.T @ search_direction) > abs(
            self._wolfe_c2 * nabla.T @ search_direction
        ):
            step *= 0.99
            x_new = _update(x, search_direction, step)
            nabla_new = _grad(*_unpack(x_new))[0]
            # print(
            #     energy(*_unpack(x_new)),
            #     fx,
            #     fx + self._wolfe_c1 * step * nabla.T @ search_direction,
            #     abs(nabla_new.T @ search_direction),
            #     abs(self._wolfe_c2 * nabla.T @ search_direction),
            # )
            tmp[0].append(step)
            tmp[1].append(energy(*_unpack(x_new)))
            tmp[2].append(fx + step * self._wolfe_c1 * nabla.T @ search_direction)
            if i > self._wolfe_iter_max:
                print("\n\n")
                # TODO go into the details and deal with the failing cases
                import matplotlib.pyplot as plt

                fig, ax = plt.subplots()
                ax.plot(tmp[0], tmp[1], lw=1, color="black")
                ax.plot(tmp[0], tmp[2], lw=1, color="tab:red")
                plt.savefig("tmp.png", dpi=300, bbox_inches="tight")
                plt.close()
                # return 1
                raise ValueError("Wolfe failed")
            i += 1
        return step



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from magnopy._spinham._validators import (
    _spins_ordered,
    _validate_atom_index,
    _validate_unit_cell_index,
)


def _get_primary_p44(alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter=None):
    r"""
    Return the primary version of the parameter.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.

    Parameters
    ----------
    alpha : int
        Index of the first atom.
    beta : int
        Index of the second atom.
    gamma : int
        Index of the third atom.
    epsilon : int
        Index of the fourth atom.
    nu : tuple of 3 int
        Unit cell for the second atom.
    _lambda : tuple of 3 int
        Unit cell for the third atom.
    rho : tuple of 3 int
        Unit cell for the fourth atom.
    parameter : (3, 3, 3, 3) :numpy:`ndarray`, optional
        Full matrix of the parameter.

    Returns
    -------
    alpha : int
        Index of the first atom.
    beta : int
        Index of the second atom.
    gamma : int
        Index of the third atom.
    epsilon : int
        Index of the fourth atom.
    nu : tuple of 3 int
        Unit cell for the second atom.
    _lambda : tuple of 3 int
        Unit cell for the third atom.
    rho : tuple of 3 int
        Unit cell for the fourth atom.
    parameter : (3, 3, 3, 3) :numpy:`ndarray`, optional
        Full matrix of the parameter. It is returned only if ``parameter is not None``.
    """

    def _ordered(mu1, alpha1, mu2, alpha2, mu3, alpha3, mu4, alpha4):
        return (
            _spins_ordered(mu1=mu1, alpha1=alpha1, mu2=mu2, alpha2=alpha2)
            and _spins_ordered(mu1=mu2, alpha1=alpha2, mu2=mu3, alpha2=alpha3)
            and _spins_ordered(mu1=mu3, alpha1=alpha3, mu2=mu4, alpha2=alpha4)
        )

    # Case 1
    if _ordered(
        mu1=(0, 0, 0),
        alpha1=alpha,
        mu2=nu,
        alpha2=beta,
        mu3=_lambda,
        alpha3=gamma,
        mu4=rho,
        alpha4=epsilon,
    ):
        pass
    # Case 2
    elif _ordered(
        mu1=(0, 0, 0),
        alpha1=alpha,
        mu2=nu,
        alpha2=beta,
        mu3=rho,
        alpha3=epsilon,
        mu4=_lambda,
        alpha4=gamma,
    ):
        alpha, beta, gamma, epsilon = alpha, beta, epsilon, gamma
        nu, _lambda, rho = nu, rho, _lambda
        if parameter is not None:
            parameter = np.transpose(parameter, (0, 1, 3, 2))
    # Case 3
    elif _ordered(
        mu1=(0, 0, 0),
        alpha1=alpha,
        mu2=_lambda,
        alpha2=gamma,
        mu3=nu,
        alpha3=beta,
        mu4=rho,
        alpha4=epsilon,
    ):
        alpha, beta, gamma, epsilon = alpha, gamma, beta, epsilon
        nu, _lambda, rho = _lambda, nu, rho
        if parameter is not None:
            parameter = np.transpose(parameter, (0, 2, 1, 3))
    # Case 4
    elif _ordered(
        mu1=(0, 0, 0),
        alpha1=alpha,
        mu2=_lambda,
        alpha2=gamma,
        mu3=rho,
        alpha3=epsilon,
        mu4=nu,
        alpha4=beta,
    ):
        alpha, beta, gamma, epsilon = alpha, gamma, epsilon, beta
        nu, _lambda, rho = _lambda, rho, nu
        if parameter is not None:
            parameter = np.transpose(parameter, (0, 3, 1, 2))
    # Case 5
    elif _ordered(
        mu1=(0, 0, 0),
        alpha1=alpha,
        mu2=rho,
        alpha2=epsilon,
        mu3=nu,
        alpha3=beta,
        mu4=_lambda,
        alpha4=gamma,
    ):
        alpha, beta, gamma, epsilon = alpha, epsilon, beta, gamma
        nu, _lambda, rho = rho, nu, _lambda
        if parameter is not None:
            parameter = np.transpose(parameter, (0, 2, 3, 1))
    # Case 6
    elif _ordered(
        mu1=(0, 0, 0),
        alpha1=alpha,
        mu2=rho,
        alpha2=epsilon,
        mu3=_lambda,
        alpha3=gamma,
        mu4=nu,
        alpha4=beta,
    ):
        alpha, beta, gamma, epsilon = alpha, epsilon, gamma, beta
        nu, _lambda, rho = rho, _lambda, nu
        if parameter is not None:
            parameter = np.transpose(parameter, (0, 3, 2, 1))
    # Case 7
    elif _ordered(
        mu1=nu,
        alpha1=beta,
        mu2=(0, 0, 0),
        alpha2=alpha,
        mu3=_lambda,
        alpha3=gamma,
        mu4=rho,
        alpha4=epsilon,
    ):
        alpha, beta, gamma, epsilon = beta, alpha, gamma, epsilon
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (-nu1, -nu2, -nu3)
        _lambda = (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3)
        rho = (rho1 - nu1, rho2 - nu2, rho3 - nu3)
        if parameter is not None:
            parameter = np.transpose(parameter, (1, 0, 2, 3))
    # Case 8
    elif _ordered(
        mu1=nu,
        alpha1=beta,
        mu2=(0, 0, 0),
        alpha2=alpha,
        mu3=rho,
        alpha3=epsilon,
        mu4=_lambda,
        alpha4=gamma,
    ):
        alpha, beta, gamma, epsilon = beta, alpha, epsilon, gamma
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (-nu1, -nu2, -nu3)
        _lambda = (rho1 - nu1, rho2 - nu2, rho3 - nu3)
        rho = (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3)
        if parameter is not None:
            parameter = np.transpose(parameter, (1, 0, 3, 2))
    # Case 9
    elif _ordered(
        mu1=nu,
        alpha1=beta,
        mu2=_lambda,
        alpha2=gamma,
        mu3=(0, 0, 0),
        alpha3=alpha,
        mu4=rho,
        alpha4=epsilon,
    ):
        alpha, beta, gamma, epsilon = beta, gamma, alpha, epsilon
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3)
        _lambda = (-nu1, -nu2, -nu3)
        rho = (rho1 - nu1, rho2 - nu2, rho3 - nu3)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 0, 1, 3))
    # Case 10
    elif _ordered(
        mu1=nu,
        alpha1=beta,
        mu2=_lambda,
        alpha2=gamma,
        mu3=rho,
        alpha3=epsilon,
        mu4=(0, 0, 0),
        alpha4=alpha,
    ):
        alpha, beta, gamma, epsilon = beta, gamma, epsilon, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3)
        _lambda = (rho1 - nu1, rho2 - nu2, rho3 - nu3)
        rho = (-nu1, -nu2, -nu3)
        if parameter is not None:
            parameter = np.transpose(parameter, (3, 0, 1, 2))
    # Case 11
    elif _ordered(
        mu1=nu,
        alpha1=beta,
        mu2=rho,
        alpha2=epsilon,
        mu3=(0, 0, 0),
        alpha3=alpha,
        mu4=_lambda,
        alpha4=gamma,
    ):
        alpha, beta, gamma, epsilon = beta, epsilon, alpha, gamma
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (rho1 - nu1, rho2 - nu2, rho3 - nu3)
        _lambda = (-nu1, -nu2, -nu3)
        rho = (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 0, 3, 1))
    # Case 12
    elif _ordered(
        mu1=nu,
        alpha1=beta,
        mu2=rho,
        alpha2=epsilon,
        mu3=_lambda,
        alpha3=gamma,
        mu4=(0, 0, 0),
        alpha4=alpha,
    ):
        alpha, beta, gamma, epsilon = beta, epsilon, gamma, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (rho1 - nu1, rho2 - nu2, rho3 - nu3)
        _lambda = (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3)
        rho = (-nu1, -nu2, -nu3)
        if parameter is not None:
            parameter = np.transpose(parameter, (3, 0, 2, 1))
    # Case 13
    elif _ordered(
        mu1=_lambda,
        alpha1=gamma,
        mu2=(0, 0, 0),
        alpha2=alpha,
        mu3=nu,
        alpha3=beta,
        mu4=rho,
        alpha4=epsilon,
    ):
        alpha, beta, gamma, epsilon = gamma, alpha, beta, epsilon
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (-lambda1, -lambda2, -lambda3)
        _lambda = (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3)
        rho = (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3)
        if parameter is not None:
            parameter = np.transpose(parameter, (1, 2, 0, 3))
    # Case 14
    elif _ordered(
        mu1=_lambda,
        alpha1=gamma,
        mu2=(0, 0, 0),
        alpha2=alpha,
        mu3=rho,
        alpha3=epsilon,
        mu4=nu,
        alpha4=beta,
    ):
        alpha, beta, gamma, epsilon = gamma, alpha, epsilon, beta
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (-lambda1, -lambda2, -lambda3)
        _lambda = (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3)
        rho = (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3)
        if parameter is not None:
            parameter = np.transpose(parameter, (1, 3, 0, 2))
    # Case 15
    elif _ordered(
        mu1=_lambda,
        alpha1=gamma,
        mu2=nu,
        alpha2=beta,
        mu3=(0, 0, 0),
        alpha3=alpha,
        mu4=rho,
        alpha4=epsilon,
    ):
        alpha, beta, gamma, epsilon = gamma, beta, alpha, epsilon
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3)
        _lambda = (-lambda1, -lambda2, -lambda3)
        rho = (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 1, 0, 3))
    # Case 16
    elif _ordered(
        mu1=_lambda,
        alpha1=gamma,
        mu2=nu,
        alpha2=beta,
        mu3=rho,
        alpha3=epsilon,
        mu4=(0, 0, 0),
        alpha4=alpha,
    ):
        alpha, beta, gamma, epsilon = gamma, beta, epsilon, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3)
        _lambda = (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3)
        rho = (-lambda1, -lambda2, -lambda3)
        if parameter is not None:
            parameter = np.transpose(parameter, (3, 1, 0, 2))
    # Case 17
    elif _ordered(
        mu1=_lambda,
        alpha1=gamma,
        mu2=rho,
        alpha2=epsilon,
        mu3=(0, 0, 0),
        alpha3=alpha,
        mu4=nu,
        alpha4=beta,
    ):
        alpha, beta, gamma, epsilon = gamma, epsilon, alpha, beta
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3)
        _lambda = (-lambda1, -lambda2, -lambda3)
        rho = (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 3, 0, 1))
    # Case 18
    elif _ordered(
        mu1=_lambda,
        alpha1=gamma,
        mu2=rho,
        alpha2=epsilon,
        mu3=nu,
        alpha3=beta,
        mu4=(0, 0, 0),
        alpha4=alpha,
    ):
        alpha, beta, gamma, epsilon = gamma, epsilon, beta, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3)
        _lambda = (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3)
        rho = (-lambda1, -lambda2, -lambda3)
        if parameter is not None:
            parameter = np.transpose(parameter, (3, 2, 0, 1))
    # Case 19
    elif _ordered(
        mu1=rho,
        alpha1=epsilon,
        mu2=(0, 0, 0),
        alpha2=alpha,
        mu3=nu,
        alpha3=beta,
        mu4=_lambda,
        alpha4=gamma,
    ):
        alpha, beta, gamma, epsilon = epsilon, alpha, beta, gamma
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (-rho1, -rho2, -rho3)
        _lambda = (nu1 - rho1, nu2 - rho2, nu3 - rho3)
        rho = (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3)
        if parameter is not None:
            parameter = np.transpose(parameter, (1, 2, 3, 0))
    # Case 20
    elif _ordered(
        mu1=rho,
        alpha1=epsilon,
        mu2=(0, 0, 0),
        alpha2=alpha,
        mu3=_lambda,
        alpha3=gamma,
        mu4=nu,
        alpha4=beta,
    ):
        alpha, beta, gamma, epsilon = epsilon, alpha, gamma, beta
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (-rho1, -rho2, -rho3)
        _lambda = (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3)
        rho = (nu1 - rho1, nu2 - rho2, nu3 - rho3)
        if parameter is not None:
            parameter = np.transpose(parameter, (1, 3, 2, 0))
    # Case 21
    elif _ordered(
        mu1=rho,
        alpha1=epsilon,
        mu2=nu,
        alpha2=beta,
        mu3=(0, 0, 0),
        alpha3=alpha,
        mu4=_lambda,
        alpha4=gamma,
    ):
        alpha, beta, gamma, epsilon = epsilon, beta, alpha, gamma
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (nu1 - rho1, nu2 - rho2, nu3 - rho3)
        _lambda = (-rho1, -rho2, -rho3)
        rho = (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 1, 3, 0))
    # Case 22
    elif _ordered(
        mu1=rho,
        alpha1=epsilon,
        mu2=nu,
        alpha2=beta,
        mu3=_lambda,
        alpha3=gamma,
        mu4=(0, 0, 0),
        alpha4=alpha,
    ):
        alpha, beta, gamma, epsilon = epsilon, beta, gamma, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (nu1 - rho1, nu2 - rho2, nu3 - rho3)
        _lambda = (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3)
        rho = (-rho1, -rho2, -rho3)
        if parameter is not None:
            parameter = np.transpose(parameter, (3, 1, 2, 0))
    # Case 23
    elif _ordered(
        mu1=rho,
        alpha1=epsilon,
        mu2=_lambda,
        alpha2=gamma,
        mu3=(0, 0, 0),
        alpha3=alpha,
        mu4=nu,
        alpha4=beta,
    ):
        alpha, beta, gamma, epsilon = epsilon, beta, gamma, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3)
        _lambda = (-rho1, -rho2, -rho3)
        rho = (nu1 - rho1, nu2 - rho2, nu3 - rho3)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 3, 1, 0))
    # Case 24
    elif _ordered(
        mu1=rho,
        alpha1=epsilon,
        mu2=_lambda,
        alpha2=gamma,
        mu3=nu,
        alpha3=beta,
        mu4=(0, 0, 0),
        alpha4=alpha,
    ):
        alpha, beta, gamma, epsilon = epsilon, beta, gamma, alpha
        nu1, nu2, nu3 = nu
        lambda1, lambda2, lambda3 = _lambda
        rho1, rho2, rho3 = rho
        nu = (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3)
        _lambda = (nu1 - rho1, nu2 - rho2, nu3 - rho3)
        rho = (-rho1, -rho2, -rho3)
        if parameter is not None:
            parameter = np.transpose(parameter, (3, 2, 1, 0))

    if parameter is None:
        return alpha, beta, gamma, epsilon, nu, _lambda, rho

    return alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter


class _P44_iterator:
    R"""
    Iterator over the (four spins & four sites) parameters of the spin Hamiltonian.
    """

    def __init__(self, spinham) -> None:
        self.container = spinham._44
        self.mc = spinham.convention.multiple_counting
        self.length = len(self.container)
        self.index = 0

    def __next__(self):
        # Case 1
        if self.index < self.length:
            self.index += 1
            return self.container[self.index - 1]
        # Case 2
        elif self.mc and self.index < 2 * self.length:
            self.index += 1
            alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter = self.container[
                self.index - 1 - self.length
            ]
            return [
                alpha,
                beta,
                epsilon,
                gamma,
                nu,
                rho,
                _lambda,
                np.transpose(parameter, (0, 1, 3, 2)),
            ]
        # Case 3
        elif self.mc and self.index < 3 * self.length:
            self.index += 1
            alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter = self.container[
                self.index - 1 - 2 * self.length
            ]
            return [
                alpha,
                gamma,
                beta,
                epsilon,
                _lambda,
                nu,
                rho,
                np.transpose(parameter, (0, 2, 1, 3)),
            ]
        # Case 4
        elif self.mc and self.index < 4 * self.length:
            self.index += 1
            alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter = self.container[
                self.index - 1 - 3 * self.length
            ]
            return [
                alpha,
                gamma,
                epsilon,
                beta,
                _lambda,
                rho,
                nu,
                np.transpose(parameter, (0, 3, 1, 2)),
            ]
        # Case 5
        elif self.mc and self.index < 5 * self.length:
            self.index += 1
            alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter = self.container[
                self.index - 1 - 4 * self.length
            ]
            return [
                alpha,
                epsilon,
                beta,
                gamma,
                rho,
                nu,
                _lambda,
                np.transpose(parameter, (0, 2, 3, 1)),
            ]
        # Case 6
        elif self.mc and self.index < 6 * self.length:
            self.index += 1
            alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter = self.container[
                self.index - 1 - 5 * self.length
            ]
            return [
                alpha,
                epsilon,
                gamma,
                beta,
                rho,
                _lambda,
                nu,
                np.transpose(parameter, (0, 3, 2, 1)),
            ]
        # Case 7
        elif self.mc and self.index < 7 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 6 * self.length]
            return [
                beta,
                alpha,
                gamma,
                epsilon,
                (-nu1, -nu2, -nu3),
                (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3),
                (rho1 - nu1, rho2 - nu2, rho3 - nu3),
                np.transpose(parameter, (1, 0, 2, 3)),
            ]
        # Case 8
        elif self.mc and self.index < 8 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 7 * self.length]
            return [
                beta,
                alpha,
                epsilon,
                gamma,
                (-nu1, -nu2, -nu3),
                (rho1 - nu1, rho2 - nu2, rho3 - nu3),
                (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3),
                np.transpose(parameter, (1, 0, 3, 2)),
            ]
        # Case 9
        elif self.mc and self.index < 9 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 8 * self.length]
            return [
                beta,
                gamma,
                alpha,
                epsilon,
                (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3),
                (-nu1, -nu2, -nu3),
                (rho1 - nu1, rho2 - nu2, rho3 - nu3),
                np.transpose(parameter, (2, 0, 1, 3)),
            ]
        # Case 10
        elif self.mc and self.index < 10 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 9 * self.length]
            return [
                beta,
                gamma,
                epsilon,
                alpha,
                (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3),
                (rho1 - nu1, rho2 - nu2, rho3 - nu3),
                (-nu1, -nu2, -nu3),
                np.transpose(parameter, (3, 0, 1, 2)),
            ]
        # Case 11
        elif self.mc and self.index < 11 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 10 * self.length]
            return [
                beta,
                epsilon,
                alpha,
                gamma,
                (rho1 - nu1, rho2 - nu2, rho3 - nu3),
                (-nu1, -nu2, -nu3),
                (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3),
                np.transpose(parameter, (2, 0, 3, 1)),
            ]
        # Case 12
        elif self.mc and self.index < 12 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 11 * self.length]
            return [
                beta,
                epsilon,
                gamma,
                alpha,
                (rho1 - nu1, rho2 - nu2, rho3 - nu3),
                (lambda1 - nu1, lambda2 - nu2, lambda3 - nu3),
                (-nu1, -nu2, -nu3),
                np.transpose(parameter, (3, 0, 2, 1)),
            ]
        # Case 13
        elif self.mc and self.index < 13 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 12 * self.length]
            return [
                gamma,
                alpha,
                beta,
                epsilon,
                (-lambda1, -lambda2, -lambda3),
                (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3),
                (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3),
                np.transpose(parameter, (1, 2, 0, 3)),
            ]
        # Case 14
        elif self.mc and self.index < 14 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 13 * self.length]
            return [
                gamma,
                alpha,
                epsilon,
                beta,
                (-lambda1, -lambda2, -lambda3),
                (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3),
                (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3),
                np.transpose(parameter, (1, 3, 0, 2)),
            ]
        # Case 15
        elif self.mc and self.index < 15 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 14 * self.length]
            return [
                gamma,
                beta,
                alpha,
                epsilon,
                (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3),
                (-lambda1, -lambda2, -lambda3),
                (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3),
                np.transpose(parameter, (2, 1, 0, 3)),
            ]
        # Case 16
        elif self.mc and self.index < 16 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 15 * self.length]
            return [
                gamma,
                beta,
                epsilon,
                alpha,
                (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3),
                (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3),
                (-lambda1, -lambda2, -lambda3),
                np.transpose(parameter, (3, 1, 0, 2)),
            ]
        # Case 17
        elif self.mc and self.index < 17 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 16 * self.length]
            return [
                gamma,
                epsilon,
                alpha,
                beta,
                (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3),
                (-lambda1, -lambda2, -lambda3),
                (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3),
                np.transpose(parameter, (2, 3, 0, 1)),
            ]
        # Case 18
        elif self.mc and self.index < 18 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 17 * self.length]
            return [
                gamma,
                epsilon,
                beta,
                alpha,
                (rho1 - lambda1, rho2 - lambda2, rho3 - lambda3),
                (nu1 - lambda1, nu2 - lambda2, nu3 - lambda3),
                (-lambda1, -lambda2, -lambda3),
                np.transpose(parameter, (3, 2, 0, 1)),
            ]
        # Case 19
        elif self.mc and self.index < 19 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 18 * self.length]
            return [
                epsilon,
                alpha,
                beta,
                gamma,
                (-rho1, -rho2, -rho3),
                (nu1 - rho1, nu2 - rho2, nu3 - rho3),
                (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3),
                np.transpose(parameter, (1, 2, 3, 0)),
            ]
        # Case 20
        elif self.mc and self.index < 20 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 19 * self.length]
            return [
                epsilon,
                alpha,
                gamma,
                beta,
                (-rho1, -rho2, -rho3),
                (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3),
                (nu1 - rho1, nu2 - rho2, nu3 - rho3),
                np.transpose(parameter, (1, 3, 2, 0)),
            ]
        # Case 21
        elif self.mc and self.index < 21 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 20 * self.length]
            return [
                epsilon,
                beta,
                alpha,
                gamma,
                (nu1 - rho1, nu2 - rho2, nu3 - rho3),
                (-rho1, -rho2, -rho3),
                (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3),
                np.transpose(parameter, (2, 1, 3, 0)),
            ]
        # Case 22
        elif self.mc and self.index < 22 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 21 * self.length]
            return [
                epsilon,
                beta,
                gamma,
                alpha,
                (nu1 - rho1, nu2 - rho2, nu3 - rho3),
                (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3),
                (-rho1, -rho2, -rho3),
                np.transpose(parameter, (3, 1, 2, 0)),
            ]
        # Case 23
        elif self.mc and self.index < 23 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 22 * self.length]
            return [
                epsilon,
                beta,
                gamma,
                alpha,
                (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3),
                (-rho1, -rho2, -rho3),
                (nu1 - rho1, nu2 - rho2, nu3 - rho3),
                np.transpose(parameter, (2, 3, 1, 0)),
            ]
        # Case 24
        elif self.mc and self.index < 24 * self.length:
            self.index += 1
            (
                alpha,
                beta,
                gamma,
                epsilon,
                (nu1, nu2, nu3),
                (lambda1, lambda2, lambda3),
                (rho1, rho2, rho3),
                parameter,
            ) = self.container[self.index - 1 - 23 * self.length]
            return [
                epsilon,
                beta,
                gamma,
                alpha,
                (lambda1 - rho1, lambda2 - rho2, lambda3 - rho3),
                (nu1 - rho1, nu2 - rho2, nu3 - rho3),
                (-rho1, -rho2, -rho3),
                np.transpose(parameter, (3, 2, 1, 0)),
            ]

        raise StopIteration

    def __len__(self):
        return self.length * (1 + 23 * int(self.mc))

    def __iter__(self):
        return self


@property
def _p44(spinham):
    r"""
    Parameters of (four spins & four sites) term of the Hamiltonian.

    .. math::

        \boldsymbol{J}_{4,4}(\boldsymbol{r}_{\nu,\alpha\beta}, \boldsymbol{r}_{\lambda,\alpha\gamma}, \boldsymbol{r}_{\rho,\alpha\varepsilon})

    of the term

    .. math::

        C_{4,4}
        \sum_{\substack{\mu, \nu, \alpha, \beta,\\ i, j, u, v}}
        J^{ijuv}_{4,4}(\boldsymbol{r}_{\nu,\alpha\beta}, \boldsymbol{r}_{\lambda,\alpha\gamma}, \boldsymbol{r}_{\rho,\alpha\varepsilon})
        S_{\mu,\alpha}^i
        S_{\mu+\nu,\beta}^j
        S_{\mu+\lambda, \gamma}^u
        S_{\mu+\rho, \varepsilon}^v

    Returns
    -------
    parameters : iterator
        List of parameters. The list has a form of

        .. code-block:: python

            [[alpha, beta, gamma, epsilon, nu, lambda, rho, J], ...]

        where

        ``alpha`` is an index of the atom located in the (0,0,0) unit cell.

        ``beta`` is an index of the atom located in the nu unit cell.

        ``gamma`` is an index of the atom located in the lambda unit cell.

        ``epsilon`` is an index of the atom located in the rho unit cell.

        ``nu`` defines the unit cell of the second atom (beta). It is a tuple of 3
        integers.

        ``lambda`` defines the unit cell of the third atom (gamma). It is a tuple of 3
        integers.

        ``rho`` defines the unit cell of the fourth atom (gamma). It is a tuple of 3
        integers.

        ``J`` is a (3, 3, 3, 3) :numpy:`ndarray`.

    See Also
    --------
    add_44
    remove_44
    """

    return _P44_iterator(spinham)


def _add_44(
    spinham,
    alpha: int,
    beta: int,
    gamma: int,
    epsilon: int,
    nu: tuple,
    _lambda: tuple,
    rho: tuple,
    parameter,
    replace=False,
) -> None:
    r"""
    Adds a (four spins & four sites) parameter to the Hamiltonian.

    Doubles of the bonds are managed automatically (independently of the convention of the
    Hamiltonian).


    Raises
    ------
    ValueError
        If an atom already has a parameter associated with it.

    Parameters
    ----------
    alpha : int
        Index of an atom from the (0, 0, 0) unit cell.

        ``0 <= alpha < len(spinham.atoms.names)``.
    beta : int
        Index of an atom from the nu unit cell.

        ``0 <= beta < len(spinham.atoms.names)``.
    gamma : int
        Index of an atom from the _lambda unit cell.

        ``0 <= gamma < len(spinham.atoms.names)``.
    epsilon : int
        Index of an atom from the rho unit cell.

        ``0 <= epsilon < len(spinham.atoms.names)``.
    nu : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the second atom.

        .. math::

            \nu
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})
    _lambda : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the third atom.

        .. math::

            \lambda
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})
    rho : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the fourth atom.

        .. math::

            \rho
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})

    parameter : (3, 3, 3, 3) |array-like|_
        Value of the parameter (:math:`3\times3\times3\times3` matrix).
    replace : bool, default False
        Whether to replace the value of the parameter if the triplet of atoms
        ``alpha, beta, gamma, nu, lambda`` or one of its duplicates already have a
        parameter associated with it.

    See Also
    --------
    p44
    remove_44

    Notes
    -----
    If ``spinham.convention.multiple_counting`` is ``True``, then this function adds
    the bond and all its duplicates to the Hamiltonian. It will cause an ``ValueError``
    to add the duplicate of the bond after the bond is added.

    If ``spinham.convention.multiple_counting`` is ``False``, then only the primary
    version of the bond is added to the Hamiltonian.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    _validate_atom_index(index=beta, atoms=spinham.atoms)
    _validate_atom_index(index=gamma, atoms=spinham.atoms)
    _validate_atom_index(index=epsilon, atoms=spinham.atoms)
    _validate_unit_cell_index(ijk=nu)
    _validate_unit_cell_index(ijk=_lambda)
    _validate_unit_cell_index(ijk=rho)
    spinham._reset_internals()

    parameter = np.array(parameter)

    alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter = _get_primary_p44(
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        epsilon=epsilon,
        nu=nu,
        _lambda=_lambda,
        rho=rho,
        parameter=parameter,
    )

    # TD-BINARY_SEARCH

    # Try to find the place for the new one inside the list
    index = 0
    while index < len(spinham._44):
        # If already present in the model
        if spinham._44[index][:7] == [alpha, beta, gamma, epsilon, nu, _lambda, rho]:
            # Either replace
            if replace:
                spinham._44[index] = [
                    alpha,
                    beta,
                    gamma,
                    epsilon,
                    nu,
                    _lambda,
                    rho,
                    parameter,
                ]
                return
            # Or raise an error
            raise ValueError(
                f"Parameter is already set for the quartet of atoms "
                f"{alpha}, {beta} {nu}, {gamma} {_lambda}, {epsilon} {rho}. Or for their duplicate."
            )

        # If it should be inserted before current element
        if spinham._44[index][:7] > [alpha, beta, gamma, epsilon, nu, _lambda, rho]:
            spinham._44.insert(
                index, [alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter]
            )
            return

        index += 1

    # If it should be inserted at the end or at the beginning of the list
    spinham._44.append([alpha, beta, gamma, epsilon, nu, _lambda, rho, parameter])


def _remove_44(
    spinham,
    alpha: int,
    beta: int,
    gamma: int,
    epsilon: int,
    nu: tuple,
    _lambda: tuple,
    rho: tuple,
) -> None:
    r"""
    Removes a (four spins & four sites) parameter from the Hamiltonian.

    Duplicates of the bonds are managed automatically (independently of the convention of
    the Hamiltonian).

    Parameters
    ----------
    alpha : int
        Index of an atom from the (0, 0, 0) unit cell.

        ``0 <= alpha < len(spinham.atoms.names)``.
    beta : int
        Index of an atom from the nu unit cell.

        ``0 <= beta < len(spinham.atoms.names)``.
    gamma : int
        Index of an atom from the _lambda unit cell.

        ``0 <= gamma < len(spinham.atoms.names)``.
    epsilon : int
        Index of an atom from the rho unit cell.

        ``0 <= epsilon < len(spinham.atoms.names)``.
    nu : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the second atom.

        .. math::

            \nu
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})
    _lambda : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the third atom.

        .. math::

            \lambda
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})
    rho : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the fourth atom.

        .. math::

            \rho
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})

    See Also
    --------
    p44
    add_44

    Notes
    -----
    If ``spinham.convention.multiple_counting`` is ``True``, then this function removes
    all versions of the bond from the Hamiltonian.

    If ``spinham.convention.multiple_counting`` is ``False``, then this function removes
    the primary version of the given bond.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    _validate_atom_index(index=beta, atoms=spinham.atoms)
    _validate_atom_index(index=gamma, atoms=spinham.atoms)
    _validate_atom_index(index=epsilon, atoms=spinham.atoms)
    _validate_unit_cell_index(ijk=nu)
    _validate_unit_cell_index(ijk=_lambda)
    _validate_unit_cell_index(ijk=rho)

    alpha, beta, gamma, epsilon, nu, _lambda, rho = _get_primary_p44(
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        epsilon=epsilon,
        nu=nu,
        _lambda=_lambda,
        rho=rho,
    )

    # TD-BINARY_SEARCH

    for index in range(len(spinham._44)):
        # As the list is sorted, there is no point in resuming the search
        # when a larger element is found
        if spinham._44[index][:7] > [alpha, beta, gamma, epsilon, nu, _lambda, rho]:
            return

        if spinham._44[index][:7] == [alpha, beta, gamma, epsilon, nu, _lambda, rho]:
            del spinham._44[index]
            spinham._reset_internals()
            return



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from magnopy._spinham._convention import Convention

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class Energy:
    r"""
    Ground state energy of the given spin Hamiltonian.

    This class is optimized for the computation of the energy for any spin
    directions for the given Hamiltonian.

    If the spin Hamiltonian is modified, then a new instance of the energy class
    should be created from it.

    Parameters
    ----------
    spinham : py:class:.SpinHamiltonian`
        Spin Hamiltonian for the calculation of energy.
    """

    def __init__(self, spinham):
        initial_convention = spinham.convention

        magnopy_convention = initial_convention.get_modified(
            spin_normalized=False, multiple_counting=False
        )

        spinham.convention = magnopy_convention

        self.spins = np.array(spinham.magnetic_atoms.spins, dtype=float)

        ########################################################################
        #                               One spin                               #
        ########################################################################

        self.J_1 = np.zeros((spinham.M, 3), dtype=float)

        for atom, parameter in spinham.p1:
            alpha = spinham.index_map[atom]

            self.J_1[alpha] += spinham.convention.c1 * parameter

        ########################################################################
        #                               Two spins                              #
        ########################################################################

        self.J_21 = np.zeros((spinham.M, 3, 3), dtype=float)

        for atom, parameter in spinham.p21:
            alpha = spinham.index_map[atom]

            self.J_21[alpha] += spinham.convention.c21 * parameter

        self.J_22 = {}

        for atom1, atom2, _, parameter in spinham.p22:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]

            if (alpha, beta) not in self.J_22:
                self.J_22[(alpha, beta)] = np.zeros((3, 3), dtype=float)

            self.J_22[(alpha, beta)] += spinham.convention.c22 * parameter

        ########################################################################
        #                              Three spins                             #
        ########################################################################

        self.J_31 = np.zeros((spinham.M, 3, 3, 3), dtype=float)

        for atom, parameter in spinham.p31:
            alpha = spinham.index_map[atom]

            self.J_31[alpha] += spinham.convention.c31 * parameter

        self.J_32 = {}

        for atom1, atom2, _, parameter in spinham.p32:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]

            if (alpha, beta) not in self.J_32:
                self.J_32[(alpha, beta)] = np.zeros((3, 3, 3), dtype=float)

            self.J_32[(alpha, beta)] += spinham.convention.c32 * parameter

        self.J_33 = {}

        for atom1, atom2, atom3, _, _, parameter in spinham.p33:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]
            gamma = spinham.index_map[atom3]

            if (alpha, beta, gamma) not in self.J_33:
                self.J_33[(alpha, beta, gamma)] = np.zeros((3, 3, 3), dtype=float)

            self.J_33[(alpha, beta, gamma)] += spinham.convention.c33 * parameter

        ########################################################################
        #                              Four spins                              #
        ########################################################################

        self.J_41 = np.zeros((spinham.M, 3, 3, 3, 3), dtype=float)

        for atom, parameter in spinham.p41:
            alpha = spinham.index_map[atom]

            self.J_41[alpha] += spinham.convention.c41 * parameter

        self.J_421 = {}

        for atom1, atom2, _, parameter in spinham.p421:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]

            if (alpha, beta) not in self.J_421:
                self.J_421[(alpha, beta)] = np.zeros((3, 3, 3, 3), dtype=float)

            self.J_421[(alpha, beta)] += spinham.convention.c421 * parameter

        self.J_422 = {}

        for atom1, atom2, _, parameter in spinham.p422:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]

            if (alpha, beta) not in self.J_422:
                self.J_422[(alpha, beta)] = np.zeros((3, 3, 3, 3), dtype=float)

            self.J_422[(alpha, beta)] += spinham.convention.c422 * parameter

        self.J_43 = {}

        for atom1, atom2, atom3, _, _, parameter in spinham.p43:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]
            gamma = spinham.index_map[atom3]

            if (alpha, beta, gamma) not in self.J_43:
                self.J_43[(alpha, beta, gamma)] = np.zeros((3, 3, 3, 3), dtype=float)

            self.J_43[(alpha, beta, gamma)] += spinham.convention.c43 * parameter

        self.J_44 = {}

        for atom1, atom2, atom3, atom4, _, _, _, parameter in spinham.p44:
            alpha = spinham.index_map[atom1]
            beta = spinham.index_map[atom2]
            gamma = spinham.index_map[atom3]
            epsilon = spinham.index_map[atom4]

            if (alpha, beta, gamma, epsilon) not in self.J_44:
                self.J_44[(alpha, beta, gamma, epsilon)] = np.zeros(
                    (3, 3, 3, 3), dtype=float
                )

            self.J_44[(alpha, beta, gamma, epsilon)] += (
                spinham.convention.c44 * parameter
            )

        spinham.convention = initial_convention

    def E_0(self, spin_directions) -> float:
        r"""

        Parameters
        ----------
        spin_directions : (M, 3) |array-like|_
            Directions of spin vectors. Only directions of vectors are used,
            modulus is ignored. ``M`` is the amount of magnetic atoms in the
            Hamiltonian. The order of spin directions is the same as the order
            of magnetic atoms in ``spinham.magnetic_atoms.spins``.

        Returns
        -------
        E_0 : float
            Classic energy of state with ``spin_directions``.
        """

        spin_directions = np.array(spin_directions, dtype=float)
        spin_directions /= np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
        spins = spin_directions * self.spins[:, np.newaxis]

        energy = 0

        energy += np.diag(self.J_1 @ spins.T).sum()

        energy += np.einsum("mij,mi,mj->m", self.J_21, spins, spins).sum()

        energy += np.einsum("miju,mi,mj,mu->m", self.J_31, spins, spins, spins).sum()

        energy += np.einsum(
            "mijuv,mi,mj,mu,mv->m", self.J_41, spins, spins, spins, spins
        ).sum()

        for alpha, beta in self.J_22:
            energy += spins[alpha] @ self.J_22[(alpha, beta)] @ spins[beta]

        for alpha, beta in self.J_32:
            energy += np.einsum(
                "iju,i,j,u",
                self.J_32[(alpha, beta)],
                spins[alpha],
                spins[alpha],
                spins[beta],
            )

        for alpha, beta in self.J_421:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_421[(alpha, beta)],
                spins[alpha],
                spins[alpha],
                spins[alpha],
                spins[beta],
            )

        for alpha, beta in self.J_422:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_422[(alpha, beta)],
                spins[alpha],
                spins[alpha],
                spins[beta],
                spins[beta],
            )

        for alpha, beta, gamma in self.J_33:
            energy += np.einsum(
                "iju,i,j,u",
                self.J_33[(alpha, beta, gamma)],
                spins[alpha],
                spins[beta],
                spins[gamma],
            )

        for alpha, beta, gamma in self.J_43:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_43[(alpha, beta, gamma)],
                spins[alpha],
                spins[alpha],
                spins[beta],
                spins[gamma],
            )

        for alpha, beta, gamma, epsilon in self.J_44:
            energy += np.einsum(
                "ijuv,i,j,u,v",
                self.J_44[(alpha, beta, gamma, epsilon)],
                spins[alpha],
                spins[beta],
                spins[gamma],
                spins[epsilon],
            )

        return float(energy)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
from wulfric.geometry import absolute_to_relative

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def output_k_resolved(
    data,
    data_headers=None,
    output_filename=None,
    kpoints=None,
    relative=False,
    rcell=None,
    flat_indices=None,
    digits=4,
    scientific_notation=True,
):
    r"""
    Outputs any k-resolved data.

    Parameters
    ----------
    data : (N, M) |array-like|_
        Some k-resolved data. N (:math:`\ge 1`) is the amount of kpoints. M is the
        number of data modes/entries.
    data_headers : (N, ) list of str, optional
        Header for the data columns. ``[f"data {i+1} for i in range(len(data))]`` by
        default.
    output_filename : str, optional
        Name of the file for saving the data. If ``None``, then return a list of lines.
    kpoints : (N, 3) |array-like|_, optional
        List of the kpoints. ``kpoints[i]`` correspond to the ``data[i]``.
    relative : bool, default False
        If ``relative == True``, then given ``kpoints`` are interpreted as relative to
        the reciprocal unit cell. Otherwise they are interpreted as absolute coordinates.
    rcell : (3, 3) |array-like|_, optional
        Reiprocal unit cell. Rows are interpreted as vectors, columns as Cartesian
        coordinates.
    flat_indices : (N, 3) |array-like|_, optional
        kpoints converted to the list of floats according to some rule. Typically used
        for plotting the band structure. If ``None`` provided, then computed automatically,
        based on the distance of the given ``kpoints`` (whether absolute or relative).
    digits : int, default 4
        Number of digits after the comma for the data elements.
    scientific_notation : bool, default True
        Whether to use scientific notation for the data elements.

    Returns
    -------
    lines : str
        Only returned if ``output_filename is None``. Use ``print("\n".join(lines))``
        to output the results to the standard output stream.

    Notes
    -----
    By default kpoints are interpreted as given in absolute coordinates in the
    reciprocal space.

    .. doctest::

        >>> import magnopy.io as mio
        >>> omegas = [[1, 2], [1.2, 2.3]]
        >>> headers = ["mode 1", "mode 2"]
        >>> kpoints = [[0, 0, 0], [0.1, 0.2, 0.3]]
        >>> lines = mio.output_k_resolved(data=omegas, data_headers=headers, kpoints=kpoints)
        >>> print("\n".join(lines))
        #   flat index         mode 1         mode 2          k_x          k_y          k_z
            0.00000000     1.0000e+00     2.0000e+00   0.00000000   0.00000000   0.00000000
            0.37416574     1.2000e+00     2.3000e+00   0.10000000   0.20000000   0.30000000

    One can tell it to interpret the kpoints as given in relative coordinates to some
    reciprocal cell

    .. doctest::

        >>> lines = mio.output_k_resolved(data=omegas, data_headers=headers, kpoints=kpoints, relative=True)
        >>> print("\n".join(lines))
        #   flat index         mode 1         mode 2         k_b1         k_b2         k_b3
            0.00000000     1.0000e+00     2.0000e+00   0.00000000   0.00000000   0.00000000
            0.37416574     1.2000e+00     2.3000e+00   0.10000000   0.20000000   0.30000000

    If the reciprocal cell is provided, then the kpoints are converter either way and
    both are printed

    .. doctest::

        >>> lines = mio.output_k_resolved(data=omegas, data_headers=headers, kpoints=kpoints, relative=True, rcell=[[1, 0, 0], [0, 2, 0], [0, 0, 3]])
        >>> print("\n".join(lines))
        #   flat index         mode 1         mode 2         k_b1         k_b2         k_b3          k_x          k_y          k_z
            0.00000000     1.0000e+00     2.0000e+00   0.00000000   0.00000000   0.00000000   0.00000000   0.00000000   0.00000000
            0.37416574     1.2000e+00     2.3000e+00   0.10000000   0.20000000   0.30000000   0.10000000   0.40000000   0.90000000
        >>> lines = mio.output_k_resolved(data=omegas, data_headers=headers, kpoints=kpoints, relative=False, rcell=[[1, 0, 0], [0, 2, 0], [0, 0, 3]])
        >>> print("\n".join(lines))
        #   flat index         mode 1         mode 2         k_b1         k_b2         k_b3          k_x          k_y          k_z
            0.00000000     1.0000e+00     2.0000e+00   0.00000000   0.00000000   0.00000000   0.00000000   0.00000000   0.00000000
            0.37416574     1.2000e+00     2.3000e+00   0.10000000   0.10000000   0.10000000   0.10000000   0.20000000   0.30000000

    One can control the number of digits in data

    .. doctest::

        >>> lines = mio.output_k_resolved(data=omegas, data_headers=headers, kpoints=kpoints, digits=6)
        >>> print("\n".join(lines))
        #   flat index           mode 1           mode 2          k_x          k_y          k_z
            0.00000000     1.000000e+00     2.000000e+00   0.00000000   0.00000000   0.00000000
            0.37416574     1.200000e+00     2.300000e+00   0.10000000   0.20000000   0.30000000

    and whether the scientific notation should be used

    .. doctest::

        >>> lines = mio.output_k_resolved(data=omegas, data_headers=headers, kpoints=kpoints, scientific_notation=False)
        >>> print("\n".join(lines))
        #   flat index    mode 1    mode 2          k_x          k_y          k_z
            0.00000000    1.0000    2.0000   0.00000000   0.00000000   0.00000000
            0.37416574    1.2000    2.3000   0.10000000   0.20000000   0.30000000

    """

    # Prepare format for the data elements
    chars = digits + 1 + 1 + 3
    if scientific_notation:
        chars += 1 + 1 + 3
    chars = max(chars, 8)

    # Format string for data elements
    if scientific_notation:
        fmt = f">{chars}.{digits}e"
    else:
        fmt = f">{chars}.{digits}f"

    # Format string for kpoints
    kp_fmt = ">12.8f"

    # Prepare the header
    header = ["#"]
    if flat_indices is not None or kpoints is not None:
        header.append(f"{'flat index':>12}")
    header = header + [f"{tmp:>{chars}}" for tmp in data_headers]

    if kpoints is not None:
        if rcell is not None:
            header.append(
                " ".join([f"{f'k_{comp}':>12}" for comp in ["b1", "b2", "b3"]])
            )
            header.append(" ".join([f"{f'k_{comp}':>12}" for comp in "xyz"]))
        elif relative:
            header.append(
                " ".join([f"{f'k_{comp}':>12}" for comp in ["b1", "b2", "b3"]])
            )
        else:
            header.append(" ".join([f"{f'k_{comp}':>12}" for comp in "xyz"]))

    header = " ".join(header)

    # Output header
    if output_filename is not None:
        f = open(output_filename, "w")
        f.write(header + "\n")
    else:
        lines = [header]

    # Output data for each kpoint
    for i in range(len(data)):
        line = [" "]

        if flat_indices is not None:
            line.append(f"{flat_indices[i]:{kp_fmt}}")
        elif kpoints is not None:
            if i == 0:
                line.append(f"{0.0:{kp_fmt}}")
            else:
                line.append(
                    f"{np.linalg.norm(np.array(kpoints[i], dtype=float) - kpoints[i-1]):{kp_fmt}}"
                )

        if isinstance(data[i], Iterable):
            for entry in data[i]:
                line.append(f"{entry:{fmt}}")
        else:
            line.append(f"{data[i]:{fmt}}")

        if kpoints is not None:
            if rcell is not None:
                if relative:
                    k_vec = kpoints[i] @ np.array(rcell)
                    line.append(
                        " ".join([f"{kpoints[i][comp]:{kp_fmt}}" for comp in range(3)])
                    )
                    line.append(
                        " ".join([f"{k_vec[comp]:{kp_fmt}}" for comp in range(3)])
                    )
                else:
                    k_vec_rel = absolute_to_relative(vector=kpoints[i], basis=rcell)
                    line.append(
                        " ".join([f"{k_vec_rel[comp]:{kp_fmt}}" for comp in range(3)])
                    )
                    line.append(
                        " ".join([f"{kpoints[i][comp]:{kp_fmt}}" for comp in range(3)])
                    )
            else:
                line.append(
                    " ".join([f"{kpoints[i][comp]:{kp_fmt}}" for comp in range(3)])
                )

        line = " ".join(line)
        if output_filename is not None:
            f.write(line + "\n")
        else:
            lines.append(line)

    if output_filename is not None:
        f.close()
    else:
        return lines


def plot_k_resolved(data, kp=None, output_filename=None, ylabel=None):
    r"""
    Plot some k-resolved data.

    If only the ``data`` are given, then an index of the omegas is used for abscissa (x
    axis).

    Parameters
    ----------
    data : (N, M) |array-like|_
        Some k-resolved data. N (:math:`\ge 1`) is the amount of kpoints. M is the
        number of data modes/entries.
    kp : :py:class:`wulfric.Kpoints`, optional.
        Instance of the :py:class:`wulfric.Kpoints` class. It should be the same
        instance that were used in the preparation of the ``data``.
    output_filename : str, optional
        Name of the file for saving the image. If ``None``, then the graph would be
        opened in the interactive matplotlib window.
    ylabel : str, optional
        Label for the ordinate (y axis).
    """

    data = np.array(data).T

    fig, ax = plt.subplots()

    if len(data.shape) == 2:
        for entry in data:
            if kp is not None:
                ax.plot(kp.flatten_points(), entry, lw=1, color="#A47864")
            else:
                ax.plot(entry, lw=1, color="#A47864")
                ax.set_xlim(0, len(entry))
    else:
        if kp is not None:
            ax.plot(kp.flatten_points(), data, lw=1, color="#A47864")
        else:
            ax.plot(data, lw=1, color="#A47864")
            ax.set_xlim(0, len(data))

    if kp is not None:
        ax.set_xticks(kp.ticks(), kp.labels, fontsize=13)
        ax.set_xlim(kp.ticks()[0], kp.ticks()[-1])
        ax.vlines(
            kp.ticks(),
            0,
            1,
            lw=0.5,
            color="grey",
            ls="dashed",
            zorder=0,
            transform=ax.get_xaxis_transform(),
        )

    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=15)

    ax.hlines(
        0,
        0,
        1,
        lw=0.5,
        color="grey",
        linestyle="dashed",
        transform=ax.get_yaxis_transform(),
    )

    if output_filename is not None:
        plt.savefig(output_filename, dpi=400, bbox_inches="tight")
    else:
        plt.show()

    plt.close()


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging

import numpy as np

_logger = logging.getLogger(__name__)

__all__ = []


def _cubic_interpolation(alpha_l, alpha_h, f_l, f_h, fp_l, fp_h):
    r"""
    Computes the minimum of a cubic interpolation for the function f(alpha) with
    values f_l, f_h and derivatives fp_l, fp_h at the points alpha_l, alpha_h.

    Parameters
    ----------
    alpha_l : float
        Lower bound of the interval.
    alpha_h : float
        Upper bound of the interval.
    f_l : float
        Value of the function at alpha_l.
    f_h : float
        Value of the function at alpha_h.
    fp_l : float
        Derivative of the function at alpha_l.
    fp_h : float
        Derivative of the function at alpha_h.

    Returns
    -------
    alpha_min : float
        Position of the minimum of the cubic interpolation for the function f(alpha).
    """

    d_1 = fp_l + fp_h - 3 * (f_l - f_h) / (alpha_l - alpha_h)
    # # print(d_1, f_l, f_h, alpha_l, alpha_h, fp_l, fp_h)
    if d_1**2 - fp_l * fp_h < 0:
        return alpha_h
    d_2 = np.sign(alpha_h - alpha_l) * np.sqrt(d_1**2 - fp_l * fp_h)

    return alpha_h - (alpha_h - alpha_l) * (fp_h + d_2 - d_1) / (fp_h - fp_l + 2 * d_2)


def _zoom(
    x_k,
    true_variables_k,
    search_direction,
    func,
    grad,
    update,
    func_0,
    grad_0,
    alpha_lo,
    alpha_hi,
    c1,
    c2,
):
    r"""
    Zoom function for the line search with strong Wolfe conditions.

    Parameters
    ----------
    x_k : :numpy:`ndarray`
        Current parameterization of the true variables. Length of the array depends
        on the ground state sub-type. The call ``update(true_variables_k, x_k)``
        has to be valid.
    true_variables_k : ...
        Current true variables. Form and type depends on the ground state sub-type.
        Calls of ``func(true_variables_k)``, ``grad(true_variables_k)`` have to be valid.
    search_direction : :numpy:`ndarray`
        Search direction. Same shape and size as ``x_k``.
    func : callable
        Energy function. The call ``func(true_variables_k)`` has to be valid.
    grad : callable
        Gradient of the energy function. The call ``grad(true_variables_k)`` has
        to be valid.
    update : callable
        Function that update the true variables to the new state defined by the
        :math:`\boldsymbol{x}` vector. The call ``update(true_variables_k, x_k)``
        has to be valid.
    func_0 : float
        Energy at the current state: ``func(true_variables_k)``.
    grad_0 : :numpy:`ndarray`
        Gradient of the energy at the current state: ``grad(true_variables_k)``.
    alpha_lo : float
        Lower bound of the step length.
    alpha_hi : float
        Upper bound of the step length.
    c1 : float
        Parameter for the Wolfe conditions.
    c2 : float
        Parameter for the Wolfe conditions.
    """

    f_0 = func_0
    fp_0 = grad_0 @ search_direction

    true_variables_k_lo = update(true_variables_k, x_k + alpha_lo * search_direction)
    true_variables_k_hi = update(true_variables_k, x_k + alpha_hi * search_direction)

    f_lo = func(true_variables_k_lo)
    fp_lo = grad(true_variables_k_lo) @ search_direction
    f_hi = func(true_variables_k_hi)
    fp_hi = grad(true_variables_k_hi) @ search_direction
    f_j_min = None
    trial_steps = 0
    while True:
        if abs(alpha_hi - alpha_lo) < np.finfo(float).eps:
            return alpha_lo
        # print("in zoom")
        alpha_j = _cubic_interpolation(
            alpha_l=alpha_lo,
            alpha_h=alpha_hi,
            f_l=f_lo,
            f_h=f_hi,
            fp_l=fp_lo,
            fp_h=fp_hi,
        )

        true_variables_k_j = update(true_variables_k, x_k + alpha_j * search_direction)

        f_j = func(true_variables_k_j)
        if f_j_min is None:
            f_j_min = f_j
        else:
            if f_j < f_j_min:
                trial_steps = 0
                f_j_min = f_j
            elif trial_steps >= 10:
                return alpha_j
            else:
                trial_steps += 1

        # # print("zoom", alpha_lo, alpha_hi, alpha_j, f_j, f_lo, f_hi)

        if f_j > f_0 + c1 * alpha_j * fp_0 or f_j >= f_lo:
            alpha_hi = alpha_j
            f_hi = f_j
            fp_hi = grad(true_variables_k_j) @ search_direction
        else:
            fp_j = grad(true_variables_k_j) @ search_direction

            if abs(fp_j) <= -c2 * fp_0:
                return alpha_j

            if fp_j * (alpha_hi - alpha_lo) >= 0:
                alpha_hi = alpha_lo
                f_hi = f_lo
                fp_hi = fp_lo

            alpha_lo = alpha_j
            f_lo = f_j
            fp_lo = fp_j


def _line_search(
    x_k,
    true_variables_k,
    search_direction,
    func,
    grad,
    update,
    func_0=None,
    grad_0=None,
    c1=1e-4,
    c2=0.9,
    max_iterations=10000,
    alpha_max=3.0,
):
    r"""
    Computes step length via line search with strong Wolfe conditions.

    Parameters
    ----------
    x_k : :numpy:`ndarray`
        Current parameterization of the true variables. Length of the array depends
        on the ground state sub-type. The call ``update(true_variables_k, x_k)``
        has to be valid.
    true_variables_k : ...
        Current true variables. Form and type depends on the ground state sub-type.
        Calls of ``func(true_variables_k)``, ``grad(true_variables_k)`` have to be valid.
    search_direction : :numpy:`ndarray`
        Search direction. Same shape and size as ``x_k``.
    func : callable
        Energy function. The call ``func(true_variables_k)`` has to be valid.
    grad : callable
        Gradient of the energy function. The call ``grad(true_variables_k)`` has
        to be valid.
    update : callable
        Function that update the true variables to the new state defined by the
        :math:`\boldsymbol{x}` vector. The call ``update(true_variables_k, x_k)``
        has to be valid.
    func_0 : float, optional
        Energy at the current state: ``func(true_variables_k)``.
    grad_0 : :numpy:`ndarray`, optional
        Gradient of the energy at the current state: ``grad(true_variables_k)``.
    c1 : float, default 1e-4
        Parameter for the Wolfe conditions.
    c2 : float, default 0.9
        Parameter for the Wolfe conditions.
    max_iterations : int, default 10000
        Maximum number of iterations.
    alpha_max: float, default 3.0
        Maximum step length.
    """

    if func_0 is None:
        func_0 = func(true_variables_k)

    if grad_0 is None:
        grad_0 = grad(true_variables_k)

    # Add pre-check of alpha = 1

    alpha_im1 = 0
    alpha_i = 1
    f_0 = func_0
    fp_0 = grad_0 @ search_direction
    f_im1 = f_0
    fp_im1 = fp_0

    for i in range(1, max_iterations):
        x_kpi = x_k + alpha_i * search_direction
        true_variables_k_i = update(true_variables_k, x_kpi)
        f_i = func(true_variables_k_i)

        if f_i > f_0 + c1 * alpha_i * fp_0 or (i > 1 and f_i >= f_im1):
            return _zoom(
                x_k=x_k,
                true_variables_k=true_variables_k,
                search_direction=search_direction,
                func=func,
                grad=grad,
                update=update,
                func_0=func_0,
                grad_0=grad_0,
                alpha_lo=alpha_im1,
                alpha_hi=alpha_i,
                c1=c1,
                c2=c2,
            )

        fp_i = grad(true_variables_k_i) @ search_direction

        if abs(fp_i) <= -c2 * fp_0:
            return alpha_i

        if fp_i >= 0:
            return _zoom(
                x_k=x_k,
                true_variables_k=true_variables_k,
                search_direction=search_direction,
                func=func,
                grad=grad,
                update=update,
                func_0=func_0,
                grad_0=grad_0,
                alpha_lo=alpha_i,
                alpha_hi=alpha_im1,
                c1=c1,
                c2=c2,
            )

        if abs(alpha_i - alpha_im1) < np.finfo(float).eps:
            return alpha_i
        # print("in line search")
        alpha_ip1 = _cubic_interpolation(
            alpha_l=alpha_im1,
            alpha_h=alpha_i,
            f_l=f_im1,
            f_h=f_i,
            fp_l=fp_im1,
            fp_h=fp_i,
        )

        if alpha_ip1 > alpha_max:
            _logger.warning(
                f"Step length exceeded maximum value of {alpha_max}. "
                "Returning maximum step length."
            )
            return alpha_max
        if alpha_ip1 < alpha_i:
            _logger.warning("Step length decreased. Keeping the previous step length.")

        alpha_im1 = alpha_i
        f_im1 = f_i
        fp_im1 = fp_i

        alpha_i = alpha_i

    raise ValueError(f"Line search failed after {max_iterations} iterations.")



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def _validate_atom_index(index, atoms) -> None:
    r"""
    Validate that the atom index is in agreement with the amount of atoms

    Parameter
    ---------
    index
        Potential index of an atom.
    atoms : dict
        Dictionary with the atoms. This function relies on ``atoms["names"]``.

    Raises
    ------
    TypeError
        If ``index`` is not an integer.
    ValueError
        If ``index`` is out of range.
    """

    if not isinstance(index, int):
        raise TypeError(
            f"Only integers are supported as atom indices, "
            f"got '{type(index)}' from '{index}'"
        )

    if not 0 <= index < len(atoms["names"]):
        raise ValueError(
            "Index should be greater or equal to 0 and less than "
            f"{len(atoms['names'])}', got {index}."
        )


def _validate_unit_cell_index(ijk) -> None:
    r"""
    Validate that ijk can specify unit cell.

    Parameters
    ----------
    ijk
        Potential index of the unit cell.

    Raises
    ------
    TypeError
        If ``ijk`` is not a ``tuple``.
    TypeError
        If either ``i``, ``j`` or ``k`` is not an ``int``.
    """

    if not isinstance(ijk, tuple):
        raise TypeError(f"Unit cell index has to be a tuple, got '{type(ijk)}'")

    if not isinstance(ijk[0], int):
        raise TypeError(
            f"First element of the unit cell index is not an 'int', got "
            f"{type(ijk[0])} from '{ijk[0]}'"
        )

    if not isinstance(ijk[1], int):
        raise TypeError(
            f"Second element of the unit cell index is not an 'int', got "
            f"{type(ijk[1])} from '{ijk[1]}'"
        )

    if not isinstance(ijk[2], int):
        raise TypeError(
            f"Third element of the unit cell index is not an 'int', got "
            f"{type(ijk[2])} from '{ijk[2]}'"
        )


def _spins_ordered(mu1, alpha1, mu2, alpha2) -> bool:
    r"""
    Compare two spins based on their positions.

    For the definition of comparison see
    :ref:`user-guide_theory-behind_multiple-counting`.

    Parameters
    ----------
    mu1 : tuple of 3 int
    alpha1 : int
    mu2 : tuple of 3 int
    alpha2 : int

    Returns
    -------
    result : bool
    """

    i1, j1, k1 = mu1
    i2, j2, k2 = mu2

    i = i2 - i1
    j = j2 - j1
    k = k2 - k1

    if (
        i > 0
        or (i == 0 and j > 0)
        or (i == 0 and j == 0 and k > 0)
        or (i == 0 and j == 0 and k == 0 and alpha1 < alpha2)
    ):
        return True

    return False



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


__version__ = "0.1.3"
__doclink__ = "magnopy.org"
__release_date__ = "22 May 2025"

from . import constants, io, scenarios
from ._diagonalization import *
from ._energy import *
from ._exceptions import *
from ._local_rf import *
from ._lswt import *
from ._package_info import *
from ._parallelization import *
from ._spinham import *
